#ifndef CAMERA_PROFILE_H
#define CAMERA_PROFILE_H

#include "guitypes.h"
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include "clb.h"

/** 
 * Class describing the intrinsic parameters of a camera.
 * A CameraProfile contains a matrix representing the 
 * intrinsic and any distortion parameters of a camera.
 *
 * The name, if set properly can be used to automatically 
 * associate the camera properties with a true image when
 * capturing (see the CaptureDialog class).  The naming
 * convention is "CAMERANAME-widthxheight".
 */
class CameraProfile{
 public:
  char name[256];
  Matrix intrinsic;
  Matrix distortion;
  void init(const char * n){
    strncpy(name,n,256);
    intrinsic=Matrix::eye(3);
    distortion=Matrix(5,1);
    distortion.setAll(0);
  }
  CameraProfile(){
    init(0);
  }
  CameraProfile(const char * n){
    init(n);
  }
  CameraProfile(const char * n,double * vals){
    init(n);
    intrinsic(0,0)=vals[0];
    intrinsic(1,1)=vals[1];
    intrinsic(0,2)=vals[2];
    intrinsic(1,2)=vals[3];
    distortion[0]=vals[4];
    distortion[1]=vals[5];
    distortion[2]=vals[6];
    distortion[3]=vals[7];
    distortion[4]=vals[9];
  }
  CameraProfile(const char * n,double fx,double fy,double px,double py){
    init(n);
    intrinsic(0,0)=fx;
    intrinsic(1,1)=fy;
    intrinsic(0,2)=px;
    intrinsic(1,2)=py;
  }    
  void setName(const char * n){
    strncpy(name,n,256);
  }
};

/**
 * Container class for several camera profiles.
 * The CameraProfiles class contains several CameraProfile 's,
 * and also maintains a currentProfile.  Images incoming
 * from the camera (or file) are assumed to have come from
 * a camera represented by the currentProfile. 
 */
class CameraProfiles{
 public:
  CameraProfile ** profiles;
  int nprofiles;
  int maxprofiles;
  int currentProfile;
  CameraProfiles(){
    maxprofiles=16;
    profiles=new CameraProfile* [maxprofiles];
    nprofiles=0;
    currentProfile=0;
  }
  ~CameraProfiles(){
    for(int i=0;i<nprofiles;i++){
      delete profiles[i];
    }
    delete profiles;
    nprofiles=0;
    maxprofiles=0;
  }
  void add(CameraProfile * profile){
    if(nprofiles>=maxprofiles){
      int newmaxprofiles=maxprofiles+16;
      CameraProfile ** newprofiles = new CameraProfile *[newmaxprofiles];
      memset(newprofiles,0,sizeof(CameraProfiles*)*newmaxprofiles);
      memcpy(newprofiles,profiles,sizeof(CameraProfile*)*maxprofiles);
      delete profiles;
      maxprofiles=newmaxprofiles;
      profiles=newprofiles;
    }
    profiles[nprofiles]=profile;
    nprofiles++;
  }
  void remove(int i){
    if(i<0||i>=nprofiles){
      return;
    }
    printf("removing profile %d\n",i);
    nprofiles--;
    delete profiles[i];
    profiles[i]=0;
    for(;i<nprofiles;i++){
      profiles[i]=profiles[i+1];
    }
    if(currentProfile>=nprofiles)
      currentProfile--;
    if(currentProfile<0)
      currentProfile=0;
  }
  void print(){
    printf("vvvvvvvvvvvvv\n");
    for(int i=0;i<nprofiles;i++){
      printf("%d %s\n",i,profiles[i]->name);
    }
  }
  void swap(int i,int j){
    CameraProfile * back=profiles[i];
    profiles[i]=profiles[j];
    profiles[j]=back;
    if(currentProfile==i)currentProfile=j;
    else if(currentProfile==j)currentProfile=i;
  }
  CameraProfile * operator[](int i){
    if(i<0 || i>=nprofiles){
      printf("cameraProfiles index out of bounds\n");
      exit(0);
    }
    return profiles[i];
  }
  void save(char * directory){
    char filename[2048];

    DIR * dir = opendir(directory);
    if(!dir){
      printf("error loading camera profiles from %s\n",directory);
      return;
    }
    struct dirent * de;    
    while(de=readdir(dir)){
      int lastdot=strlen(de->d_name)-4;
      snprintf(filename,2048,"%s/%s",directory,de->d_name);
      if((strstr(de->d_name,".clb")-de->d_name)==lastdot){
	struct stat buf;
	stat(filename,&buf);
	if(S_ISREG(buf.st_mode)){
	  printf("removing %s\n",filename);
	  ::remove(filename);
	}
      }
    }
    closedir(dir);

    for(int i=0;i<nprofiles;i++){
      snprintf(filename,2048,"%s/%s.clb",directory,profiles[i]->name);
      FILE * file = fopen(filename,"wb");
      if(file){
	CalibInfo ci;
	int flags=HAS_DISTORTION|HAS_INTRINSIC;
	writeCLB(file,flags,profiles[i]->intrinsic,Matrix(0,0),Matrix(0,0),
		 profiles[i]->distortion,ci);
	fclose(file);
      }
      else{
	printf("error saving camera profile to: %s\n",filename);
      }
    }
    printf("saved %d camera profiles\n",nprofiles);
  }
  void load(char * directory){
    char filename[2048];
    char nprof=0;
    DIR * dir = opendir(directory);
    if(!dir){
      printf("error loading camera profiles from %s\n",directory);
      return;
    }
    struct dirent * de;    
    while(de=readdir(dir)){
      int lastdot=strlen(de->d_name)-4;
      snprintf(filename,2048,"%s/%s",directory,de->d_name);
      if((strstr(de->d_name,".clb")-de->d_name)==lastdot){
	char name[256];
	memcpy(name,de->d_name,strlen(de->d_name));
	name[lastdot]=0;
	CameraProfile * profile = new CameraProfile(name);
	CalibInfo ci;
	Matrix extr(4,4);
	Matrix proj(3,4);
	FILE * file = fopen(filename,"rb");
	if(file){
	  int flags=readCLB(file,profile->intrinsic,proj,extr,
			    profile->distortion,ci);
	  add(profile);
	  fclose(file);
	}
	else{
	  delete profile;
	  printf("error reading camera profile file %s\n",filename);
	}
	nprof++;
      }
    }
    closedir(dir);
    printf("Done loading camera profiles, found %d\n",nprof);
  }
  //Set the profile to that having the name in string.
  //returns 1 if the profile is found, and 0 otherwise
  int setCurrentProfile(const char * string){
    for(int i=0;i<nprofiles;i++){
      if(!strcmp(string,profiles[i]->name)){
	currentProfile=i;
	return 1;
      }
    }
    return 0;
  }
  int getCurrentParameters(Matrix & intrinsic,
			    Matrix & distortion){
    distortion=Matrix(5,1);
    intrinsic=Matrix::eye(3);
    
    if(currentProfile>=0 &&
       currentProfile<nprofiles){
      distortion=profiles[currentProfile]->distortion;
      intrinsic=profiles[currentProfile]->intrinsic;
      return 1;
    }
    return 0;
    /*
    double fx=363.0017;//413.1795329561925;
    double fy=363.5357;//413.3604660994160;
    double px=152.360809036324;
    double py=113.112125563360;

    distortion[0]=-0.40743818922655;
    distortion[1]= 0.53393818146118;
    distortion[2]= 0.00036239062064;
    distortion[3]=-0.00380086091997;
    distortion[4]=0;
	
    intrinsic(0,0)=fx;
    intrinsic(1,1)=fy;
    intrinsic(0,2)=px;
    intrinsic(1,2)=py;
    */
  }
  
  /** Find CameraProfile that matches the parameters.
   *  Locate and return the camera profile that matches the given parameters.
   *  Returns NULL if one cannot be found.
   * \param intrinsic the intrinsic parameters to match.
   * \param distortion the distortion parameters to match.
   */
  CameraProfile * findCameraProfile(Matrix & intrinsic,
				    Matrix & distortion){
    printf("locating profile in %d\n",nprofiles);
    for(int i=0;i<nprofiles;i++){
      Matrix diff1=intrinsic-profiles[i]->intrinsic;
      Matrix diff2=distortion-profiles[i]->distortion;
      if(diff1.dot(diff1)<0.01 &&
	 diff2.dot(diff2)<0.001){
	printf("found it\n");
	return profiles[i];
      }
    }
    return 0;
  }
};

#endif
