#ifndef GL_CANVAS_H
#define GL_CANVAS_H

#include <wx/wx.h>
#include <sys/time.h>
#include "image.h"
#include "matrix.h"
#include "newseg.h"
#include "imagedata.h"
#include <wx/glcanvas.h>
#include "guitypes.h"
#include "chart.h"
#include "timer.h"

/**An abstract class for recieving mouse events from the GLCanvas.
 * This abstract class allows recieving of mouse events from a GLCanvas.
 * The drawGL function is called by the GLCanvas, allowing for the
 * GLMouseListener to draw client dependent data.
 */
class GLMouseListener{
 public:
  virtual void drawGL(int w,int h)=0;
  virtual void mouse(wxMouseEvent & event,int w,int h)=0;
};

///A simple rectangle class.
class Wrect{
 public:
  double x;
  double y;
  double width;
  double height;
  Wrect(){
    x=y=width=height=0;
  }
  Wrect(double _x,double _y,double w,
	double h){
    x=_x;
    y=_y;
    width=w;
    height=h;
  }
};

///The available GLCanvas modes
typedef enum {
  GL_CANVAS_IMAGE,
  GL_CANVAS_MODEL,
  GL_CANVAS_TEXTURE,
  GL_CANVAS_GETTEX,
  NUM_GLCANVASMODES} GLCanvasMode;

/** GLCanvas extends wxGLCanvas and allows for drawing to a openGL window.
 * This class extends wxGLCanvas and allows for
 * drawing to a openGL window.  The canvas has 
 * several different states (see GLCanvasMode),
 * each of which handles events (mouse/keys) and drawing.
*/
class GLCanvas: public wxGLCanvas{
  ///Current CalibInfo
  CalibInfo * calibInfo; 

  ///Last Image Data
  ImageData * lastImageData;

  GLuint tex;
  
  ///Texture for the target
  GLuint targetTex;

  ///texture for the model
  GLuint modelTex;       
  int imagew;
  int imageh;
  int hasAlpha;
  double xsize;
  double ysize;
  int w,h; //viewport width and height
  GLMouseListener * reciever;
    
  GLCanvasMode lastMode;
  GLCanvasMode mode;

  //TEXTURE mode
  double xpos;
  double ypos;
  double centroidx;
  double centroidy;
  double rot;
  double selpt[2];
  double curpt[2];

  //MODEL Mode
  wxMenu * modelModeMenu;
  Vec3d cpos;
  Vec3d crot;
  float speed[3];
  float speedrx,speedry;
  unsigned char keys[256];

 public:
  //GETTEX mode
  int index;
  char where[1024];
  int texw[3];
  int texh[3];
  int nreig[3];
  double * intrinsics;
  double * extrinsics;
  double projection[16];
  Image8 * images;
  unsigned char * texImages;
  unsigned char * occluded;
  int nimages;
  //Lookup params
  int lutw;
  int luth;
  int flags;
    
 public:
  DECLARE_DYNAMIC_CLASS(GLCanvas)
    GLCanvas(wxWindow * parent=0,int *attribs=0):
    wxGLCanvas(parent,-1,wxDefaultPosition,
	       wxDefaultSize,0,wxT("GLCanvas"),
	       attribs){

    modelModeMenu=new wxMenu();
    modelModeMenu->AppendCheckItem(ID_Wireframe,wxT("Wireframe"));
    modelModeMenu->AppendCheckItem(ID_DrawNormals,wxT("Draw Normals"));
    modelModeMenu->AppendCheckItem(ID_DrawGrid,wxT("Draw Grid"));
    modelModeMenu->AppendCheckItem(ID_DrawCameras,wxT("Draw Cameras"));
    modelModeMenu->AppendCheckItem(ID_DrawTexture,wxT("Draw Texture"));
    modelModeMenu->AppendCheckItem(ID_ProjectTexture,wxT("Project Texture"));

    tex=0;
    reciever=0;
    imagew=4;
    imageh=3;
    calibInfo=0;
    xsize=ysize=1;
      
    cpos=Vec3d(0,0,-5);
    crot=Vec3d(0,0,0);
    centroidx=centroidy=0;
    xpos=ypos=rot=0;
    hasAlpha=0;

    speedrx=speedry=0.1;
    speed[0]=speed[1]=speed[2]=0.01;
    
    lastImageData=0;
    
#ifdef WX_2_6
    //Need this because the default is huge when proper aspecting 2D
    SetMinSize(wxSize(64,48));
#endif
  }
  void UpdateTexture(Image8 & im){
    if(tex){
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,tex);
      imagew=im.width;
      imageh=im.height;
      im.initTextureNoResize(&xsize,&ysize);
      Refresh();
    }
    if(calibInfo){
      delete calibInfo;
      calibInfo=0;
    }
    hasAlpha=0;
  }
  void UpdateTexture(ImageData & imageData){
    if(tex){
      hasAlpha=0;
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,tex);
      imagew=imageData.image.width;
      imageh=imageData.image.height;

      if(imageData.flags&HAS_MASK){
	Image8 im(imageData.image.w,imageData.image.h,
		  imageData.image.nchannels+1);
	for(int i=0;i<imageData.image.h;i++){
	  for(int j=0;j<imageData.image.w;j++){
	    int k=0;
	    for(;k<imageData.image.nchannels;k++){
	      im(j,i,k)=imageData.image(j,i,k);
	    }
	    im(j,i,k)=imageData.mask(j,i)*255;
	  }
	}
	im.initTextureNoResize(&xsize,&ysize);
	hasAlpha=1;
      }
      else
	imageData.image.initTextureNoResize(&xsize,&ysize);
      Refresh();
    }
    if(calibInfo){
      delete calibInfo;
      calibInfo=0;
    }
    if(imageData.calibInfo){
      calibInfo=imageData.calibInfo->copy();
    }
    lastImageData=&imageData;
  }
  void OnPaint(wxPaintEvent & event){
    static float rot=0;
    rot+=1;
    wxPaintDC dc(this);
	
#ifndef __WXMOTIF__
    if (!GetContext()) return;
#endif
    SetCurrent();

    GetSize(&w,&h);
    glViewport(0,0,w,h);


    if(tex==0){
      glEnable(GL_TEXTURE_2D);
      glGenTextures(1,&tex);
      glGenTextures(1,&targetTex);
      glGenTextures(1,&modelTex);
      glBindTexture(GL_TEXTURE_2D,targetTex);
      Image8 im;
      wxString fname =  wxGetHomeDir()+wxT("/.capsys/target.tga");
      
      //fname.Printf("%s/.capsys//target.tga",wxGetHomeDir().ToAscii());//getenv("HOME"));
      im.readTGA(fname.mb_str());
      im.initTexture();
      glDisable(GL_TEXTURE_2D);
    }
    switch(mode){
    case GL_CANVAS_IMAGE:
      imageModePaint(event);
      break;
    case GL_CANVAS_MODEL:
      modelModePaint(event);
      break;	
    case GL_CANVAS_TEXTURE:
      textureModePaint(event);
      break;
    case GL_CANVAS_GETTEX:
      gettexModePaint(event);
      break;
    }
    if(reciever){
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glOrtho(0,w,h,0,-100,100);
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      reciever->drawGL(w,h);
    }
    SwapBuffers();
  }
  void drawChecks();
  void imageModePaint(wxPaintEvent & event);
  void gettexModePaint(wxPaintEvent & event);
  void modelModePaint(wxPaintEvent & event);
  void textureModePaint(wxPaintEvent & event);

  void drawCalibInfo(){
    glColor3f(1.0,0.0,0.0);
    for(int i=0;i<calibInfo->nlines;i++){
      int biggest=calibInfo->sizes[calibInfo->lines[i][0]];
      if(calibInfo->sizes[calibInfo->lines[i][1]]>biggest)
	biggest=calibInfo->sizes[calibInfo->lines[i][1]];
      if(calibInfo->sizes[calibInfo->lines[i][2]]>biggest)
	biggest=calibInfo->sizes[calibInfo->lines[i][2]];
      if(calibInfo->sizes[calibInfo->lines[i][3]]>biggest)
	biggest=calibInfo->sizes[calibInfo->lines[i][3]];
      int middle=(int)(0.75*biggest);
	    
      for(int j=0;j<4;j++){
	if(calibInfo->sizes[calibInfo->lines[i][j]]>middle){
	  glColor3f(0.8,1.0,0.8);
	  glPointSize(6);
	}
	else{
	  glColor3f(0.8,1.0,0.8);
	  glPointSize(3);
	}
	glBegin(GL_POINTS);
	glVertex2d(calibInfo->centers[calibInfo->lines[i][j]],
		   calibInfo->centers[calibInfo->lines[i][j]+calibInfo->ncenters]);
	glEnd();
		
      }
      glLineWidth(1);
      glColor3f(0.0,0.0,1.0);
      glBegin(GL_LINES);
      glVertex2d(calibInfo->centers[calibInfo->lines[i][0]],
		 calibInfo->centers[calibInfo->lines[i][0]+calibInfo->ncenters]);
      glVertex2d(calibInfo->centers[calibInfo->lines[i][3]],
		 calibInfo->centers[calibInfo->lines[i][3]+calibInfo->ncenters]);
      glEnd();
	    
      if(!calibInfo->labels){
	/*drawString(calibInfo->centers[calibInfo->lines[i][3]],
		   calibInfo->centers[calibInfo->lines[i][3]+calibInfo->ncenters],
		   "    %d\n",i);*/
      }
      else {
	/*drawString(calibInfo->centers[calibInfo->lines[i][3]],
		   calibInfo->centers[calibInfo->lines[i][3]+calibInfo->ncenters],
		   "   %d,%d",i,calibInfo->labels[i]);*/
	/*		if(calibInfo->labels[i]!=calibInfo->labels2[i])
			glColor3f(0.1,1.0,0.1);
			drawString(calibInfo->centers[calibInfo->lines[i][3]],
			calibInfo->centers[calibInfo->lines[i][3]+calibInfo->ncenters],
			"   %d,%d,%d",i,calibInfo->labels[i],calibInfo->labels2[i]);*/
      }
      /*
	printf("%d:%d,%d:%d,%d:%d,%d:%d\n",
	calibInfo->lines[i][0],calibInfo->sizes[calibInfo->lines[i][0]],
	calibInfo->lines[i][1],calibInfo->sizes[calibInfo->lines[i][1]],
	calibInfo->lines[i][2],calibInfo->sizes[calibInfo->lines[i][2]],
	calibInfo->lines[i][3],calibInfo->sizes[calibInfo->lines[i][3]]);*/
    }
	
    glPointSize(2);
    glBegin(GL_POINTS);
    glColor3f(0.0,0.0,1.0);
    glVertex2f(calibInfo->center[0],calibInfo->center[1]);
    glEnd();
    glColor3f(1.0,1.0,1.0);
    /*drawString(0,20,"Residual:%f",
      calibInfo->residual);*/
  }
  void SetMode(GLCanvasMode m){
    lastMode=mode;
    mode=m;
    Refresh();
        
    //GETTEX params
    index=0;
    if(mode==GL_CANVAS_TEXTURE){
      setChartFlags();
    }
  }
  void OnEraseBackground(wxEraseEvent & event){
  }
  void OnSize(wxSizeEvent & event){
  }
  void OnEnterWindow(wxMouseEvent & event){
  }
  void OnKeyDown(wxKeyEvent & event){
    int key=event.GetKeyCode();
    if(key<255){
      keys[key]=1;
    }
    printf("canvas on key down %d %c\n",event.GetKeyCode(),event.GetKeyCode());
    Refresh();
  }
  void OnKeyUp(wxKeyEvent & event){
    printf("canvas on key up\n");
    int key=event.GetKeyCode();
    if(key<255){
      keys[key]=0;
    }
  }
  void RegisterListener(GLMouseListener * comm){
    reciever=comm;
  }
  void UnRegisterListener(GLMouseListener * comm){
    reciever=NULL;
  }
  void OnMouse(wxMouseEvent & event){
    switch(mode){
    case GL_CANVAS_IMAGE:
      imageModeOnMouse(event);
      break;
    case GL_CANVAS_TEXTURE:
      textureModeOnMouse(event);
      break;
    case GL_CANVAS_MODEL:
      modelModeOnMouse(event);
      break;
    }

    if(reciever){
      reciever->mouse(event,w,h);
      Refresh();
    }
  }
  void OnIdle(wxIdleEvent & event);
  void imageModeOnMouse(wxMouseEvent & event);
  void textureModeOnMouse(wxMouseEvent & event);
  void modelModeOnMouse(wxMouseEvent & event);
  void modelModeMove();
  void modelModeMenuCheck(int id,const bool check){
    modelModeMenu->Check(id,check);
  }
  void setChartFlag(Chart & chart);
  void setChartFlags();

#ifdef WIN32 //Win32
  void BuggyCheckBox(int id){
    modelModeMenu->Check(id,!modelModeMenu->IsChecked(id));
  }
  void OnDrawNormals(wxCommandEvent & event){
    BuggyCheckBox(ID_DrawNormals);
  }
  void OnWireframe(wxCommandEvent & event){
    BuggyCheckBox(ID_Wireframe);
  }
  void OnDrawGrid(wxCommandEvent & event){
    BuggyCheckBox(ID_DrawGrid);
  }
  void OnDrawCameras(wxCommandEvent & event){
    BuggyCheckBox(ID_DrawCameras);
  }
  void OnDrawTexture(wxCommandEvent & event){
    BuggyCheckBox(ID_DrawTexture);
  }
  void OnProjectTexture(wxCommandEvent & event){
    BuggyCheckBox(ID_ProjectTexture);
  }
#endif
 protected:
  DECLARE_EVENT_TABLE();
};

#endif
