#ifndef PLUGIN_H
#define PLUGIN_H
#include <wx/wx.h>
#include <wx/string.h>
#ifdef WIN32
#include "dlfcn_win.h"
#else
#include <dlfcn.h>
#endif
#include "chart.h"
#include "plugin_defines.h"

#ifdef WIN32
#define PLUG_EXTENSION ".dll"
#else
#define PLUG_EXTENSION ".so"
#endif

/*
  Plugins:
  There are two main types of plugins, those that operate directly
  on the 3D geometry, and those that operate on the texture coordinates.
  The former are the GeometryPlugins and the latter are TexturePlugins.

  Every plugin must provide more than 3 functions, of which
  must contain:
  1)const char getName(void ): a function returning the name of 
  the plugin.  This name will be used to identify the controls for the
  plugin (be it a button, or a tabbed page)
  2)const char getInfo(void ): a function returning the info of
  the plugin.  Currently this is used just to set the tooltip for 
  the control of the object(be it a button, or panel).
  3)int getType(): The name type of the function, which will be 
  discussed below.

  A few notes before continuing.  In general, the plugins provide a function
  that gets called when the plugin is activated (ie. apply or create)
  and another function (getPanel) that returns a wxPanel with the options 
  for the plug.
  
  The return value of apply or create function is used to interface with
  the application, and should be 0 on success.  In some cases,
  if getPanel function returns 0, the plugin will be placed differently in
  the application. This function currently only gets called once, but there is
  no gaurantee, so plugins should take all precautions (ie: if(panel)return panel)
  Furthermore, the appilcation uses the ClientData and ToolTip of the
  returned panel, so may not be set by the plugin.
  
  There are currently two types of GeometryPlugins.  The values these
  plugins getType function should return are given below.  This is followed 
  by other details:
  1)GEOMPLUGIN_CREATE_TYPE: 
     Create the geometry from the images
     i)provides a function "int (* create)(char * dir, char * file,int nimages,
                                          char * geomFile,
		                          float sx,float sy,float sz,
		                          float tx,float ty,float tz);"
      -The create function gives the director to load the images from,
       the name of the image file (if image file is "file" then the
       calibration is "file.clb").  The output geometry should be written
       to geomFile.  The sx,sy,sz and tx,ty,tz are the scaling and translation
       of the unit cube to contain the object.
      -The create function will first be called with all NULL arguments, asking
       the plugin for any requirements.  Currently, the only return value is
       GEOMPLUGIN_NEEDS_BOUNDING_BOX_BIT, suggesting that the main application
       should first get the bounding box, then recall the apply function.
       If the function returns 0 on the NULL argument call, then no bounding
       box will be supplied, and the remaining arguments should be ignored.

     ii)provides a function "int (* getPanel)(wxWindow * parent)"
      -This function gives the panel of options for the method.
      -This panel will be added as a tab on a tabbed pane.  The name
       of the tab will be set with name from getName(), and the tooltip
       will be set with the getInfo() function.
      -The plugin may return 0 for this, suggesting that no options are
       required.
  2)GEOMPLUGIN_OPERATION_TYPE: Perform an operation on the geometry 
     i) Provides "int apply(Chart ** charts,int * ncharts);"
      -The function allows the plugin to delete the current charts,
       and allocate a new set of charts.
      -The function returns a bit mask of:
         GEOMPLUGIN_APPLY_AGAIN_BIT : the function will be called
                                      immediately after its return.
         GEOMPLUGIN_RESTORE_BIT     : the geometry will be restored
                                      immediately after the function returns.
         GEOMPLUGIN_MERGE_BIT       : the geometry will be merged
                                      after the function returns.
         The reason for these is so that if the geometry has been split
         (can be tested with (*ncharts)>1) then the function can ask
         the application to restore or merge the charts, and can also ask
         the application to call the function again, after the charts
         have been merged with the APPLY_AGAIN_BIT.
     ii) Provides "int (*getPanel)(wxWindow * parent)"
       -Gives a panel of options for the plugin. 
       -If this is non-null, the plugin will get its own tab on a tabbed pane.
       -If this function returns null, the plugin has no options, and will 
        appear on the last tabbed pane as a button.

  There are three types of TexturePlugins.  The values these
  plugins getType() function should return are given below.  Each
  texture plugin must also provide the following functions
  I)int apply(Chart ** charts,int * ncharts);
    -the function that gets called to initiate the plugin
    -always return 0
  II)wxPanel * getPanel(wxWindow *);
    -the function that returns the options panel for the plugin.  In
     each case, this may return null.
  1)TEXPLUGIN_SPLIT_TYPE:
    -The plugin is given an entry in a comboBox, and when selected, the
     panel from getPanel will be shown.  If getPanel() returns 0, then 
     a default panel is provided. 
  2)TEXPLUGIN_MAP_TYPE
    -The plugin is given a radioButton.  The getPanel() function IS ignored!.
     No options allowed.  In the future this may be supported in a similar
     manner as 1).
  3)TEXPLUGIN_OTHER_TYPE
    -If getPanel() returns NULL, the plugin is given a button in the "Others"
     tab.  If the getPanel() returns non-null, the plugin is given its own
     tab, named with getName() and the panel is placed in the tab.
 
 */


wxArrayString findPlugins(const char * basedir, const char * type);

class Plugin{
 protected:
  void * userData;
  void * handle;
 public:
  int (*getType)();
  const char * (* getInfo)(void );
  const char * (* getName)(void );
  Plugin(){
    userData=0;
    handle=0;
  }
  void setUserData(void * v){
    userData=v;
  }
  void  * getUserData(){
    return userData;
  }
  virtual int open(const char * path){
    char * error=0;
    //allow reopening a different file, close automatically
    if(handle)
      dlclose(handle);
          
    handle=dlopen((char *)path,RTLD_LAZY);
    if ((error = dlerror()) != NULL)  {
      fprintf (stderr, "%s\n", error);
      return 0;
    }
    return handle>0;
  }
  virtual int load(){
    char * error;
    if(!handle)return 0;
    getName=(const char*(*)()) dlsym(handle,"getName");
    if ((error = dlerror()) != NULL)  {
      fprintf (stderr, "%s\n", error);
      return 0;
    }
    getInfo=(const char*(*)()) dlsym(handle,"getInfo");
    if ((error = dlerror()) != NULL)  {
      fprintf (stderr, "%s\n", error);
      return 0;
    }
    getType=(int (*)())dlsym(handle,"getType");
    if ((error = dlerror()) != NULL)  {
      fprintf (stderr, "%s\n", error);
      return 0;
    }
    return 1;
  }
  virtual int loaded(){
    return handle!=0;
  }
  virtual int close(){
    if(handle){
      dlclose(handle);
      handle=0;
    }
  }
  virtual ~Plugin(){
    close();
    userData=0;
  }
};



class GeometryPlugin : public Plugin{
 public:
  union{
    int (* apply )(Chart ** charts,int * ncharts);
    int (* create)(char * dir, char * file,int nimages,
		   char * geomFile,
		   float sx,float sy,float sz,
		   float tx,float ty,float tz);
  };
  wxPanel * (* getPanel)(wxWindow *);
  virtual int load(){
    char * error;
    if(!Plugin::load())return 0;
    if(getType()==GEOMPLUGIN_OPERATION_TYPE){
      apply=(int (*)(Chart ** charts,int * ncharts))
	dlsym(handle,"apply");
      if((error=dlerror())!=NULL){
	fprintf(stderr,"%s\n",error);
	return 0;
      }
    }
    else{
      create=(int (*)(char *, char *, int,char *,
		      float,float,float,float,float,float))
	dlsym(handle,"create");
      if((error=dlerror())!=NULL){
	fprintf(stderr,"%s\n",error);
	return 0;
      }
    }
    getPanel =(wxPanel * (*)(wxWindow *))dlsym(handle,"getPanel");
    if ((error = dlerror()) != NULL)  {
      fprintf (stderr, "%s\n", error);
      return 0;
    } 
    return 1;
  }
};



class TexturePlugin : public Plugin{
 public:
  int (* apply )(Chart ** charts,int * ncharts);
  wxPanel * (* getPanel)(wxWindow *);
  virtual int load(){
    char * error;
    if(!Plugin::load())return 0;
    apply=(int (*)(Chart ** ,int *))dlsym(handle,"apply");
    if ((error = dlerror()) != NULL)  {
      fprintf (stderr, "%s\n", error);
      return 0;
    }
    getPanel=(wxPanel * (*)(wxWindow *))dlsym(handle,"getPanel");
    if ((error = dlerror()) != NULL)  {
      fprintf (stderr, "%s\n", error);
      return 0;
    }
    return 1;
  }
};


#endif
