#include "defmesh.h"
#include "image.h"
#include "matrix.h"
#include "pqt.h"
#include "voxbin.h"
#include <stdarg.h>

#ifdef USING_GLEE
#include <GLee.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glext.h>
#endif

//VoxBin tvb(100);  
//VoxBin po;
//DefMesh target;
//PointOctree po;

//Some globals, that shouldn't be.
//vector<int> *   modified=0;
//vector<int> * lastModified=0;


//#define CHECK 1
int vertexNeighbors(HalfEdge * he,HalfEdge ** neighbors,int max);

#define MAX_LECHAUD_REMESH 4
int lechaudMod[MAX_LECHAUD_REMESH];

int printErr(char * str)
{
  int err=0;
  if((err=glGetError())){
    printf("GLERROR (%s) : %d ",str,err);
    switch(err){
    case GL_INVALID_ENUM:
      printf("invalid enum\n");break;
    case GL_INVALID_VALUE:
      printf("invalid value\n");break;
    case GL_INVALID_OPERATION:
      printf("invalid operation\n");break;
    case GL_STACK_OVERFLOW:
      printf("stack overflow\n");break;
    case GL_STACK_UNDERFLOW:
      printf("stack underflow\n");break;
    case GL_OUT_OF_MEMORY:
      printf("out of mem\n");break;
    case GL_TABLE_TOO_LARGE:
      printf("table too large\n");break;
    }
    return err;
  }
  return 0;
}

inline void printDebug(const char * fmt,...)
{
#ifdef PD
  va_list ap;
  va_start(ap,fmt);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
#endif
}

/** Constructor for Defmesh
 * \param sigma_ the "invariant"(lechaud) of the mesh.
 * \sa DefMesh::~DefMesh
 */
DefMesh::DefMesh(double sigma_)
{
  triangleTexWidth=0;
  desiredTriangleWidth=5;
 
  sigma=sigma_;
  nvert=0;
  ntri=0;

  maxtri=100;
  maxvert=300;
  
  tri = new Face[maxtri];
  vert = new Vec3f[maxvert];
  norm = new Vec3f[maxvert];

  lastnconn=1;
  maxvalence=0;
  useTexNumber=0;
  
  required_texw=-1;
  required_texh=-1;
}
/** Destructor for Defmesh
 */
DefMesh::~DefMesh()
{
  for(int i=0;i<ntri;i++){
    HalfEdge * finger=tri[i].edge;
    int j=0;
    for(int j=0;finger && j<3;j++){
      HalfEdge * prev=finger;
      finger=finger->next;
      prev->next=0;
      prev->opp=0;
      delete prev;
    }
  }
  delete [] tri;
  delete [] vert;
  delete [] norm;
  vert=0;
  norm=0;
  tri=0;
  maxtri=0;
  maxvert=0;
  ntri=0;
  nvert=0;
}

/** Generate a sphere by subdivision
    \param rad the radius of the sphere
 */
void DefMesh::genSphere(double rad)
{
  ensureVert(4);
  ensureTri(4);
  
  nvert=4;
  
  double z=rad/3;
  double sc=sqrt(rad*rad-z*z);
  vert[0]=Vec3f(0,0,rad);
  vert[1]=Vec3f(sc*cos(2.0*M_PI/3.0),
		sc*sin(2.0*M_PI/3.0),-z);
  vert[2]=Vec3f(sc*cos(2*2.0*M_PI/3.0),
		sc*sin(2*2.0*M_PI/3.0),-z);
  vert[3]=Vec3f(sc*cos(0),
		sc*sin(0),-z);
  HalfEdge * he1,* he2,* he3;
  allocFace(&he1,&he2,&he3);
  he1->vert=2;
  he2->vert=1;
  he3->vert=3;
  allocFace(&he1,&he2,&he3);
  he1->vert=1;
  he2->vert=2;
  he3->vert=0;
  allocFace(&he1,&he2,&he3);
  he1->vert=2;
  he2->vert=3;
  he3->vert=0;
  allocFace(&he1,&he2,&he3);
  he1->vert=3;
  he2->vert=1;
  he3->vert=0;
  printDebug("diffs  %f %f %f %f %f\n",(vert[0]-vert[1]).len(),
	 (vert[1]-vert[2]).len(),
	 (vert[2]-vert[3]).len(),(vert[0]-vert[3]).len(),
	 (vert[0]-vert[2]).len());
  printDebug("vert[3].len() %f\n",vert[3].len());
  printDebug("vert[2].len() %f\n",vert[2].len());
  printDebug("vert[1].len() %f\n",vert[1].len());
  printDebug("vert[0].len() %f\n",vert[0].len());
  for(int i=0;i<4;i++){
    vert[i].normalize();
    vert[i]*=rad;
  }

  for(int i=0;i<4;i++){
    int oldntri=ntri;
    for(int j=0;j<oldntri;j++){
      Vec3f v1=vert[tri[j].edge->vert];
      Vec3f v2=vert[tri[j].edge->next->vert];
      Vec3f v3=vert[tri[j].edge->next->next->vert];
      Vec3f v1v2,v1v3,v2v3;
      v1v2=(v1+v2)*0.5;
      v1v2.normalize();
      v1v2*=rad;

      v1v3=(v1+v3)*0.5;
      v1v3.normalize();
      v1v3*=rad;

      v2v3=(v2+v3)*0.5;
      v2v3.normalize();
      v2v3*=rad;
      
      int v1i=tri[j].edge->vert;
      int v2i=tri[j].edge->next->vert;
      int v3i=tri[j].edge->next->next->vert;
    
      int v1v2i=-1;
      int v2v3i=-1;
      int v1v3i=-1;

      for(int k=0;k<nvert;k++){
	double diff=(v1v2-vert[k]).len();
	if(diff<0.00001)v1v2i=k;
	diff=(v1v3-vert[k]).len();
	if(diff<0.00001)v1v3i=k;
	diff=(v2v3-vert[k]).len();
	if(diff<0.00001)v2v3i=k;
	if(v1v2i!=-1 && v1v3i!=-1 &&
	   v2v3i!=-1)break;
      }
      if(v1v2i==-1){
	v1v2i=nvert;
	ensureVert(++nvert);
	vert[v1v2i]=v1v2;
      }
      if(v2v3i==-1){
	v2v3i=nvert;
	ensureVert(++nvert);
	vert[v2v3i]=v2v3;
      }
      if(v1v3i==-1){
	v1v3i=nvert;
	ensureVert(++nvert);
	vert[v1v3i]=v1v3;
      }
      
      tri[j].edge->vert=v1i;
      tri[j].edge->next->vert=v1v2i;
      tri[j].edge->next->next->vert=v1v3i;
      
      HalfEdge * he1,* he2,* he3;
      allocFace(&he1,&he2,&he3);
      he1->vert=v1v2i;
      he2->vert=v2i;
      he3->vert=v2v3i;

      allocFace(&he1,&he2,&he3);
      he1->vert=v1v2i;
      he2->vert=v2v3i;
      he3->vert=v1v3i;
      
      allocFace(&he1,&he2,&he3);
      he1->vert=v1v3i;
      he2->vert=v2v3i;
      he3->vert=v3i;
    }
  }
  printf("%d\n",nvert);

  for(int i=nvert;i<nvert;i++){
    double sc=0.4*fabs(sin(8*M_PI*(vert[i].z/rad)/4.0))+0.01;
    //double sc=0.9*fabs(cos(4*M_PI*(vert[i].z/rad)/4.0))+0.1;
    vert[i].x*=sc;
    vert[i].y*=sc;
  }

  
  fixAdj();
  initNormals();
}

/** Ensure there is enough vertices, reallocate the
    required amount if necessary.
    \param n The required number of vertices
    \sa DefMesh::ensureTri
 */
void DefMesh::ensureVert(int n)
{
  if(n>maxvert){
    int newmaxvert=(int)(MAX(maxvert*1.3+10,n));
    Vec3f * vec3fp = new Vec3f[newmaxvert];
    memcpy(vec3fp,vert,sizeof(Vec3f)*maxvert);
    delete [] vert;
    vert=vec3fp;
    vec3fp = new Vec3f[newmaxvert];
    memcpy(vec3fp,norm,sizeof(Vec3f)*maxvert);
    delete [] norm;
    norm=vec3fp;
    maxvert=newmaxvert;
  }
}

/** Ensure there is enough triangles, reallocate the
    required amount if necessary.
    \param n The required number of triangles
    \sa DefMesh::ensureTri
 */
void DefMesh::ensureTri(int n)
{
  if(n>maxtri){
    int newmaxtri=(int)(MAX(maxtri*1.3+10,n));
    Face * he = new Face[newmaxtri];
    memcpy(he,tri,sizeof(Face)*maxtri);
    delete [] tri;
    tri=he;
    maxtri=newmaxtri;
  }
}

/** Remove all triangles and vertices (do not release memory)
 */
void DefMesh::clear()
{
  ntri=0;
  nvert=0;
}

/** Allocate a face.
    \param he1 pointer to a pointer of the first edge
    \param he2 pointer to a pointer of the second edge
    \param he3 pointer to a pointer of the third edge
    \pre he1,he2 and he3 are all non-null.
    \post The halfEdges will be allocated, and the desired face is given
          by the face of the he1,he2,he3.  The pointers within the
          halfEdges will be set appropriately, but the verts still
          need to be set.
    \sideeffect none
 */
void DefMesh::allocFace(HalfEdge ** he1,HalfEdge ** he2,HalfEdge ** he3)
{
  (*he1)=allocHalfEdge();
  (*he2)=allocHalfEdge();
  (*he3)=allocHalfEdge();
  ensureTri(ntri+1);
  
  (*he1)->face=ntri;
  (*he2)->face=ntri;
  (*he3)->face=ntri;

  (*he1)->next=*he2;
  (*he2)->next=*he3;
  (*he3)->next=*he1;
  
  tri[ntri].flags=0;
  tri[ntri++].edge=*he1;
}

/** Allocate a halfEdge.
    \post A HalfEdge is allocated, initialized, and returned.
    \return A pointer to a new HalfEdge
    \sideeffect none
 */
HalfEdge * DefMesh::allocHalfEdge()
{
  HalfEdge * he=new HalfEdge();
  he->vert=0;
  he->opp=0;
  return he;
}

/** Free a halfEdge.
    \param he The HalfEdge to free
    \pre HalfEdge was allocated with DefMesh::allocHalfEdge
    \post Free the memory for he
    \return void
    \sideeffect none
 */
void DefMesh::freeHalfEdge(HalfEdge * he)
{
  if(he){
    delete he;
  }
}

void DefMesh::recursiveDeleteCnt(HalfEdge * he,hash_map<int,char> & vis,int label){
  if(vis.count(he->face))return;
  vis[he->face]=label;
  for(int k=0;k<3;k++){
    if(he->opp){
      recursiveDeleteCnt(he->opp,vis);
    }
    he=he->next;
  }
}

int DefMesh::recursiveDelete(HalfEdge * he,int & cnt,int & fixed){
  if(!he)return cnt;
  if(!tri[he->face].edge)return cnt;
  cnt++;
  if(cnt>2000)return -1;
  tri[he->face].edge=0;
  for(int k=0;k<3;k++){
    if(he->opp){
      if(he->opp->opp)
	he->opp->opp=0;
      if(recursiveDelete(he->opp,cnt,fixed)<0)return -1;
      he->opp=0;
    }
    else{
      fixed++;
    }
    he=he->next;
  }
  //  freeHalfEdge(he->next->next);
  //freeHalfEdge(he->next);
  //freeHalfEdge(he);
  //floating memory stuff
  return cnt;
}

/** Fix up any of the HalfEdge->opp pointers
    \post All HalfEdges should have an opposite after this.
    \return void
    \sideeffect none
 */
int DefMesh::fixAdj(){
  int ** vert2tri=new int * [nvert];
  int * nvert2tri=new int[nvert];
  int * maxvert2tri=new int[nvert];
  
  printDebug("fixing adj\n");
  for(int i=0;i<nvert;i++){
    maxvert2tri[i]=10;
    vert2tri[i]=new int[maxvert2tri[i]];
    nvert2tri[i]=0;
  }
  for(int i=0;i<ntri;i++){
    HalfEdge * finger = tri[i].edge;
    int j=0;
    for(j=0;j<3;j++){
      int v=finger->vert;
      if(nvert2tri[v]>=maxvert2tri[v]){
	int newmax=maxvert2tri[v]+10;
	int * newintptr=new int[newmax];
	memcpy(newintptr,vert2tri[v],maxvert2tri[v]*sizeof(int));
	delete vert2tri[v];
	vert2tri[v]=newintptr;
	maxvert2tri[v]=newmax;
      }
      vert2tri[v][nvert2tri[v]++]=finger->face;
      finger=finger->next;
    }
  }
  vector<HalfEdge *> multi;
  for(int i=0;i<ntri;i++){
    HalfEdge * he=tri[i].edge;
    for(int h=0;h<3;h++){
      int v=he->vert;
      if(he->opp){
	he->opp->opp=0;
	he->opp=0;
      }

      //if(he->opp){he=he->next;continue;}
      vector<HalfEdge *> possible;
      for(int k=0;k<nvert2tri[v];k++){
	int tri2=vert2tri[v][k];
	if(tri2==i)continue;
	HalfEdge * he2=tri[tri2].edge;
	int found=0;
	for(int jj=0;jj<3;jj++){
	  if(he2->next->vert==v){
	    found=1;
	    break;
	  }
	  he2=he2->next;
	}
	if(found && he2->next->vert == he->next->vert){
	  printf("same edge\n");
	}
	if(found && he2->vert==he->next->vert){
	  possible.push_back(he2);
	}
      }
      if(possible.size()==1){
	he->opp=possible[0];
	possible[0]->opp=he;
      }
      else if(possible.size()>1){
	printf("more than one choice\n");
	multi.push_back(he);
      }
      he=he->next;
    }
  }

  //again need to check possible choices
  for(int i=0;i<multi.size();i++){
    HalfEdge * he=multi[i];
    int v=he->vert;
    vector<HalfEdge *> possible;
    for(int k=0;k<nvert2tri[v];k++){
      int tri2=vert2tri[v][k];
      if(tri2==i)continue;
      HalfEdge * he2=tri[tri2].edge;
      int found=0;
      for(int jj=0;jj<3;jj++){
	if(he2->next->vert==v){
	  found=1;
	  break;
	}
	he2=he2->next;
      }
      if(found && he2->next->vert == he->next->vert){
	printf("same edge\n");
      }
      if(found && he2->vert==he->next->vert){
	possible.push_back(he2);
      }
    }
    if(possible.size()>1){
      hash_map<int,char> tvis;
      if(he->opp)he->opp->opp=0;
      he->opp=0;
      recursiveDeleteCnt(he,tvis);
      printf("tvis: %d (%d)\n",tvis.size(),ntri);
      int maxj=0;
      int max=0;
      for(int j=0;j<possible.size();j++){
	hash_map<int,char> vis;
	recursiveDeleteCnt(possible[j],vis);
	printf("%d =>%d\n",j,vis.size());
	if(vis.size()>max){
	  max=vis.size();
	  maxj=j;
	}
      }
      if(tvis.size()>ntri/4)
      {
	printf("using =>%d\n",maxj);
	if(tvis.size()==max && possible[maxj]->opp){
	  if(tvis.size()==max){
	    possible[maxj]->opp->opp=0;
	    possible[maxj]->opp=he;
	    he->opp=possible[maxj];
	  }
	}
	else if(tvis.size()==max){
	  possible[maxj]->opp=he;
	  he->opp=possible[maxj];
	}
      }
    }
    he=he->next;
  }

  //Check to see that all hes have an opp
  int noopps=0;
  vector<HalfEdge * > oppLess;
  vector<HalfEdge * > holeTri;
  for(int i=0;i<ntri;i++){
    HalfEdge * he=tri[i].edge;
    if(!he)continue;
    int ontri=0;
    for(int h=0;h<3;h++,he=he->next){
      if(!he->opp){
	printf("vert %d to %d has no opp (%f,%f,%f) to (%f,%f,%f) triangle:%d\n",
	       he->vert,he->next->vert,
	       vert[he->vert].x,vert[he->vert].y,vert[he->vert].z,
	       vert[he->next->vert].x,
	       vert[he->next->vert].y,
	       vert[he->next->vert].z,i);
	oppLess.push_back(he);
	ontri++;
	noopps++;
      }
      he=he->next;
    }
    //This triangle has no opposites
    if(ontri==3){
      holeTri.push_back(he);
    }
  }
  if(oppLess.size()){
    int fixed=0;

    if(holeTri.size()){
      for(int t=0;t<holeTri.size();t++){
	HalfEdge * hes[3]={holeTri[t],
			   holeTri[t]->next,
			   holeTri[t]->next->next};
	int fnd[]={0,0,0};
	HalfEdge * opps[3]={0,0,0};
	for(int i=0;i<oppLess.size();i++){
	  for(int k=0;k<3;k++){
	    if(oppLess[i]!=hes[k] &&
	       oppLess[i]->vert==hes[k]->vert &&
	       oppLess[i]->next->vert==hes[k]->next->vert){
	      opps[k]=oppLess[i];
	      fnd[k]++;
	    }
	  }
	}
	if(fnd[0]==1 && fnd[1]==1 && fnd[2]==1){
	  printf("found backwards facing triangle ... flipping it\n");
	  hes[0]->next=hes[2];
	  hes[2]->next=hes[1];
	  hes[1]->next=hes[0];
	  //indices changed, since the order flips
	  hes[1]->opp=opps[0];
	  hes[2]->opp=opps[1];
	  hes[0]->opp=opps[2];
	  hes[0]->opp->opp=hes[0];
	  hes[1]->opp->opp=hes[1];
	  hes[2]->opp->opp=hes[2];
	  fixed+=6;
	}
	else if(fnd[0]>1 || fnd[1]>1||fnd[2]>1){
	  printf("no idea what to do here\n");
	  printf("bailing out\n");
	  return 0;
	}
      }
    }
    if(oppLess.size()-fixed>0){
      printf("There are still %d halfedges (of %d) with no opposite \n",
	     oppLess.size()-fixed,ntri*3);
      int head=0;
      for(int i=0;i<oppLess.size();i++){
	if(!oppLess[i]->opp){
	  printf("opps: %p,%p %p,%p %p,%p\n",oppLess[i],oppLess[i]->opp,
		 oppLess[i]->next,oppLess[i]->next->opp, 
		 oppLess[i]->next->next,oppLess[i]->next->next->opp);
	  printf("calling recursive delete\n");
	  int del=0;
	  int rdelfix=0;
	  hash_map<int,char> vis;
	  recursiveDeleteCnt(oppLess[i],vis);
	  printf("cnt:%d\n",vis.size());
	  if(vis.size()<ntri/4){
	    if(recursiveDelete(oppLess[i],del,rdelfix)<0){
	      printf("deleting entire mesh (%d)\n",rdelfix);
	      return 0;
	    }
	  }
	  printf("deleted %d, %d\n",del,rdelfix);
	  fixed+=rdelfix;
	}
      }


      int nop=0;
      int bad=0;
      int badv=0;
      hash_map<int,HalfEdge * > vto;

      for(int i=0;i<ntri;i++){
	HalfEdge * he=tri[i].edge;
	if(!he)continue;
	int ontri=0;
	for(int h=0;h<3;h++,he=he->next){
	  
	  if(!he->opp){
	    vto[he->vert]=he;
	    printf("vert %d to %d has no opp (%f,%f,%f) to (%f,%f,%f) triangle:%d\n",
		   he->vert,he->next->vert,
		   vert[he->vert].x,vert[he->vert].y,vert[he->vert].z,
		   vert[he->next->vert].x,
		   vert[he->next->vert].y,
		   vert[he->next->vert].z,i);
	    nop++;
	  }
	  	  
	  else if(he->opp && he->face==he->opp->face){
	    printf("the opposite is the same face\n");
	  }
	  else if(he->opp->opp!=he){
	    bad++;
	  }
	  else if(he->next->vert !=he->opp->vert ||
	     he->vert !=he->opp->next->vert){
	    badv++;
	  }


	  if(he->opp->next->vert!=he->vert){
	    printf("err: opposite's shouldn't be opposite\n");
	    printf("     (%d,%p):%d-%d opposite: (%d,%p):%d-%d\n",
		   he->face,he,he->vert,he->next->vert,
		   he->opp->face,he->opp,he->opp->vert,he->opp->next->vert);
	  }
	  if(he->face!=i)
	    printf("err: face is not set correctly\n");
	  if(he->next->next->next!=he)
	    printf("err: inner triangle not set correctly\n");
	  if(he->next==he)
	    printf("err: he is itself\n");
	  if(he->opp==he)
	    printf("err: he opp is itself\n");
	  if(!tri[he->opp->face].edge)
	    printf("err: opp=%d of %d has been deleted\n",he->opp->face,i);
	  if(he->vert==he->next->vert)
	    printf("err: degenerate edge %d %d\n",he->vert,he->next->vert);
	  HalfEdge * finger = he;
	  int v=he->vert;
	  int attempts=0;
	  while(attempts<100){
	    if(finger->vert!=v){
	      printf("err: adjacent vertices are screwed up for vert %d\n",
		     he->vert);
	      break;
	    }
	    finger=finger->opp->next;
	    if(finger==he)break;
	    attempts++;
	  }
	  if(attempts==100){
	    printf("err: unable to finish checking neighbors\n");
	  }
	}
	he=he->next;
      }
      printf("nop :%d, bad:%d, badv:%d\n",nop,bad,badv);
      //see if we can find some loops
      hash_map<int,HalfEdge *>::iterator finger=vto.begin();
      hash_map<int,HalfEdge *>::iterator end=vto.end();
     
      while(finger!=end){
	int v=(*finger).first;
	HalfEdge * fing=((*finger).second);
	if(vto[v]){
	  vector<HalfEdge *> order;
	  
	  while(vto.count(fing->next->vert) &&
		vto[fing->next->vert] &&
		fing->next->vert !=v){
	    order.push_back(fing);
	    fing=vto[fing->next->vert];
	  }
	  order.push_back(fing);
	  printf("order.size()=%d\n",order.size());
	  if(fing->next->vert==v && order.size()>=3){
	    vector<HalfEdge *> newEdges;
	    for(int i=2;i<order.size();i++){
	      HalfEdge * e1;
	      HalfEdge * e2;
	      HalfEdge * e3;
	      allocFace(&e1,&e2,&e3);
	      e1->vert=order[0]->vert;
	      e2->vert=order[i]->vert;
	      e3->vert=order[i-1]->vert;
	      printf("face %d %d %d\n",e1->vert,e2->vert,e3->vert);
	      newEdges.push_back(e1);
	      newEdges.push_back(e2);
	      newEdges.push_back(e3);
	    }

	    for(int i=0;i<order.size();i++){
	      vto[order[i]->vert]=0;
	    }

	    //set up half edges
	    for(int i=0;i<newEdges.size();i++){
	      HalfEdge * he1=newEdges[i];
	      if(he1->opp)continue;
	      for(int k=0;k<newEdges.size();k++){
		HalfEdge * he2=newEdges[k];
		if(he2->opp)continue;
		if(he1->vert==he2->next->vert &&
		   he2->vert==he1->next->vert){
		  he1->opp=he2;
		  he2->opp=he1;
		}
	      }
	      if(he1->opp)continue;
	      for(int k=0;k<order.size();k++){
		HalfEdge * he2=order[k];
		if(he2->opp)continue;
		if(he1->vert==he2->next->vert &&
		   he2->vert==he1->next->vert){
		  he1->opp=he2;
		  he2->opp=he1;
		}
	      }
	    }
	  }
	}
	finger++;
      }
      vector<HalfEdge * > deleteList;
      hash_map<int,char> vis;
      for(int i=1;i<255;i++){
	int j=0;
	for(j=0;j<ntri;j++){
	  if(tri[j].edge && !vis.count(j)){
	    break;
	  }
	}
	int before=vis.size();
	printf("about to get count\n");
	recursiveDeleteCnt(tri[j].edge,vis,i);
	int after=vis.size();
	printf("got count\n");
	if(after-before<128){
	  int n=0,b=0;
	  printf("about to delete\n");
	  recursiveDelete(tri[j].edge,n,b);
	  printf("deleted %d\n",n);
	}

	printf("%d\n",vis.size());
	if(vis.size()==ntri)break;
      }
      printf("all done %d\n",deleteList.size());
      printf("bailing out\n");
      return 1;
    }

  }

  for(int i=0;i<nvert;i++)
    delete vert2tri[i];
  delete [] vert2tri;
  delete [] maxvert2tri;
  delete [] nvert2tri;
  printDebug("done fixing adj\n");
  return 1;
}

/** Intialize the normals of the mesh.
    \post All the vertices are fixed up
    \return void
    \sideeffect Any unused vertices will be removed
 */
void DefMesh::initNormals(){
  printDebug("initing Normals\n");
  int * vmap = new int[nvert];
  unsigned char * valid = new unsigned char[nvert];
  memset(valid,0,nvert*sizeof(char));
  memset(norm,0,sizeof(Vec3f)*(nvert));
  
  for(int i=0;i<ntri;i++){
    if(tri[i].edge==0)continue;
    int v1=tri[i].edge->vert;
    int v2=tri[i].edge->next->vert;
    int v3=tri[i].edge->next->next->vert;
    Vec3f a = vert[v2];
    Vec3f b = vert[v3];
    a.sub(vert[v1]);
    b.sub(vert[v1]);
    Vec3f c=a.cross(b);
    double len=c.normalize();
    
    Vec3f e0=vert[v2]-vert[v1];e0.normalize();
    Vec3f e1=vert[v3]-vert[v1];e1.normalize();
    
    double a0=acos(e0.dot(e1));
    
    e0=vert[v1]-vert[v2];e0.normalize();
    e1=vert[v3]-vert[v2];e1.normalize();
    
    double a1=acos(e0.dot(e1));
    
    norm[v1]+=(c*a0);
    norm[v2]+=(c*a1);
    norm[v3]+=(c*(M_PI-a0-a1));
    //avoid overflow
    if(valid[v1]==255)valid[v1]=254;
    if(valid[v2]==255)valid[v2]=254;
    if(valid[v3]==255)valid[v3]=254;
    valid[v1]++;
    valid[v2]++;
    valid[v3]++;
    //c.print();
  }
  maxvalence=0;
  int used=0;
  for(int i=0;i<nvert;i++){
    double l=(norm)[i].normalize();
    if(!valid[i]){
      vmap[i]=-1;
      continue;
    }
    else{
      maxvalence=MAX(maxvalence,valid[i]/2);
    }

    l=norm[i].len();
    if(isnan(l)||isinf(l)||l>1.0001){
#warning "FIX bad normals"
      norm[i]=Vec3f(0,0,0);
    }
    if(isnan(vert[i].x)||isinf(vert[i].x) ||
       isnan(vert[i].y)||isinf(vert[i].y) ||
       isnan(vert[i].z)||isinf(vert[i].z) ){
      printDebug("point is nan or inf \n");
    }
    norm[used]=norm[i];
    vert[used]=vert[i];
    vmap[i]=used++;
  }
  
  for(int i=0;i<ntri;i++){
    if(tri[i].edge==0)continue;
    if(vmap[tri[i].edge->vert]<0){
      printf("its negative %d %d %d\n",tri[i].edge->vert,
	     tri[i].edge->next->vert,
	     tri[i].edge->next->next->vert);
    }
    tri[i].edge->vert=vmap[tri[i].edge->vert];
    tri[i].edge->next->vert=vmap[tri[i].edge->next->vert];
    tri[i].edge->next->next->vert=vmap[tri[i].edge->next->next->vert];
  }
  //printf("used %d/%d\n",used,nvert);
  nvert=used;
  delete [] valid;
  delete [] vmap;
  printDebug("done initing normals\n");
}

/** Write the .obj file for the mesh
    \param filename the file to write the .obj file to.
    \pre filename is non null
    \post the file is written to .obj
    \return 0 on failure, 1 on success
    \sideeffect Unused triangles and unused vertices(via initNormals)
    will be removed
 */
int DefMesh::writeobj(char * filename){
  removeUnusedTri();
  initNormals();

  FILE * file = fopen(filename,"wb");
  if(!file)return 0;
  for(int i=0;i<nvert;i++){
    fprintf(file,"v %f %f %f\n",vert[i].x,vert[i].y,vert[i].z);
  }
  for(int i=0;i<ntri;i++){
    HalfEdge * he=tri[i].edge;
    if(!he)continue;
    fprintf(file,"f %d %d %d\n",he->vert+1,
	    he->next->vert+1,
	    he->next->next->vert+1);
  }

  fclose(file);
}

/** Load the .obj file
    \param filename the file to read the .obj from.
    \pre filename is non null
    \post this is initialized with the data from the file
    \return 0 on failure, 1 on success
    \sideeffect Any adjacencies will be fixed, and the
    normals will be initialized.
 */
int DefMesh::loadobj(char * filename,int center){
  FILE * file=fopen(filename,"rb");
  if(!file){
    printDebug("error opening %s\n",filename);
    return 0;
  }
  printDebug("loading object %s\n",filename);
  clear();
  char line[1024];

  vector<Vec2f> tc;
  vector<Vec3<int> > tritc;
  while(fgets(line,1023,file)){
    if(strstr(line,"v ")==line){
      ensureVert(nvert+1);
      sscanf(line+2,"%f %f %f",
	     &(vert[nvert].x),
	     &(vert[nvert].y),
	     &(vert[nvert].z));
      nvert++;
    }
    else if(strstr(line,"vt ")==line){
      //ignore for now
      Vec2f t;
      sscanf(line+3,"%f %f",&t.x,&t.y);
      tc.push_back(t);
    }
    else if(strstr(line,"f ")==line){
      ensureTri(ntri+1);
      HalfEdge * e1=allocHalfEdge();
      HalfEdge * e2=allocHalfEdge();
      HalfEdge * e3=allocHalfEdge();
      Vec3<int> tcind;
      //If there is texture coords then add them
      if(6==sscanf(line+2,"%d/%d %d/%d %d/%d",
		   &(e1->vert),&(tcind.x),
		   &(e2->vert),&(tcind.y),
		   &(e3->vert),&(tcind.z))){
	tcind.x--;
	tcind.y--;
	tcind.z--;
	tritc.push_back(tcind);
      }else
	sscanf(line+2,"%d %d %d",&(e1->vert),&(e2->vert),&(e3->vert));
      e1->next=e2;
      e2->next=e3;
      e3->next=e1;
      e1->face=ntri;
      e2->face=ntri;
      e3->face=ntri;
      tri[ntri++].edge=e1;
      e1->vert--;
      e2->vert--;
      e3->vert--;
    }
  }
  //add the tex coords
  if(tc.size() && tritc.size()==ntri){
    printf("have texture coords\n");
    for(int i=0;i<tritc.size();i++){
      for(int k=0;k<3;k++){
	tri[i].tc[k]=tc[tritc[i][k]];
      }
    }
  }
  //assume there is a one-one correspondence between vert and tc
  else if(tc.size()==nvert){
    for(int i=0;i<ntri;i++){
      HalfEdge * he=tri[i].edge;
      for(int k=0;k<3;k++,he=he->next){
	tri[i].tc[k]=tc[he->vert];
      }
    }
  }

  if(center){
    Vec3f avg(0,0,0);
    for(int i=0;i<nvert;i++){
      avg+=vert[i];
    }
    avg*=1.0/nvert;
    for(int i=0;i<nvert;i++){
      vert[i]-=avg;
    }
  }
#warning Should remove duplicate verts only if necessary
  /*
  int * map=new int[nvert];
  int nunused=0;
  memset(map,-1,sizeof(int)*nvert);
  for(int i=0;i<nvert;i++){
    if(map[i]==-1){
      for(int j=i+1;j<nvert;j++){
	double diff=(vert[i]-vert[j]).len();
	if(diff<10e-8){
	  map[j]=i;
	  nunused++;
	}
      }
    }
  }
  for(int i=0;i<ntri;i++){
    HalfEdge * he=tri[i].edge;
    for(int j=0;j<3;j++){
      if(map[he->vert]>=0){
	he->vert=map[he->vert];
      }
      he=he->next;
    }
  }
  //use the map to hold the new mapping
  int used=0;
  for(int i=0;i<nvert;i++){
    int back=map[i];
    map[i]=used;
    if(back==-1){
      used++;
    }
  }
  printf("used %d/%d\n",used,nvert);
  //fill in the holes
  for(int i=0;i<ntri;i++){
    HalfEdge * he=tri[i].edge;
    for(int j=0;j<3;j++){
      if(map[he->vert]>=0){
	he->vert=map[he->vert];
      }
      he=he->next;
    }
  }
  nvert-=nunused; 
  printf("nunused %d\n",nunused);
  */
  printDebug("done loading\n");
  if(!fixAdj())return 0;
  initNormals();
  printDebug("returning from load obj\n");
  return 1;
}

/** move the vertices of the mesh
  Still under developlment
 */
void DefMesh::move()
{
  /*if(target.nvert==0){
    printDebug("moving\n");
    for(int i=0;i<nvert;i++){
      vert[i]+=(norm[i]*0.3);
    }
    initNormals();
    }*/
  /*else{
    static int cnt=0;
    if(cnt==0){
      printDebug("initing to\n");
      Vec3f min( 10e10, 10e10, 10e10);
      Vec3f max(-10e10,-10e10,-10e10);
      for(int i=0;i<target.nvert;i++){
	min.x=MIN(min.x,target.vert[i].x);
	min.y=MIN(min.y,target.vert[i].y);
	min.z=MIN(min.z,target.vert[i].z);
     	max.x=MAX(max.x,target.vert[i].x);
	max.y=MAX(max.y,target.vert[i].y);
	max.z=MAX(max.z,target.vert[i].z);
      }
      tvb.setRanges(min.x,max.x,min.y,max.y,min.z,max.z);
      for(int i=0;i<target.nvert;i++)tvb.insert(target.vert[i]);
      for(int i=0;i<target.ntri;i++){
	HalfEdge * he1,* he2, * he3;
	he1=target.tri[i].edge;
	he2=he1->next;
	he3=he2->next;

	Vec3f e1=target.vert[he2->vert]-target.vert[he1->vert];
	Vec3f e2=target.vert[he3->vert]-target.vert[he1->vert];
	double e1len=e1.len();
	double e2len=e2.len();
	for(double a=0;a<1;a+=0.015/e1len){
	  for(double b=0;b<1;b+=0.015/e2len){
	    if(a+b<=1){
	      Vec3f v=target.vert[he1->vert];
	      v+=e1*a;
	      v+=e2*b;
	      tvb.insert(v);
	    }
	  }
	}
      }
    }
    for(int i=0;i<nvert;i++){
      if(!tvb.distanceQuery(vert[i],sigma)){
	vert[i]-=(norm[i]*0.6*sigma);
      }
      else
	printDebug("stopped moving\n");
    }
    cnt++;
    }*/
}
void DefMesh::drawTC(){
  
  glBegin(GL_TRIANGLES);
  for(int i=0;i<ntri;i++)
  {
    HalfEdge * he=tri[i].edge;
    if(!he)continue;
    glBegin(GL_TRIANGLES);
    glTexCoord2fv(tri[i].tc[0].data);
    glVertex2fv(tri[i].tc[0].data);
    glTexCoord2fv(tri[i].tc[1].data);
    glVertex2fv(tri[i].tc[1].data);
    glTexCoord2fv(tri[i].tc[2].data);
    glVertex2fv(tri[i].tc[2].data);
    glEnd();
  }
  glEnd();
}

//just a raw drawing of triangles
//notice that the TS_DEPTH_MODE flags is ignored, as it should be
void DefMesh::drawme(int flags)
{
  if(flags&(TS_USE_TC_FOR_VERT|TS_USE_VERT_FOR_TC)){
    glBegin(GL_TRIANGLES);
    for(int i=0;i<ntri;i++){
      HalfEdge * he;
      if(he=tri[i].edge){
	glNormal3fv(norm[he->vert].data);
	glTexCoord3fv(vert[he->vert].data);
	glVertex2fv(tri[i].tc[0].data);
	
	glNormal3fv(norm[he->next->vert].data);
	glTexCoord3fv(vert[he->next->vert].data);
	glVertex2fv(tri[i].tc[1].data);
		
	glNormal3fv(norm[he->next->next->vert].data);
	glTexCoord3fv(vert[he->next->next->vert].data);
	glVertex2fv(tri[i].tc[2].data);      
      }
    }
    glEnd();
  }
  else{
    for(int i=0;i<ntri;i++)
    {
      HalfEdge * he=tri[i].edge;
      if(!he)continue;
      glBegin(GL_TRIANGLES);
      glNormal3fv(norm[he->vert].data);
      glTexCoord2fv(tri[i].tc[0].data);
      glVertex3fv(vert[he->vert].data);
      glNormal3fv(norm[he->next->vert].data);
      glTexCoord2fv(tri[i].tc[1].data);
      glVertex3fv(vert[he->next->vert].data);
      glNormal3fv(norm[he->next->next->vert].data);
      glTexCoord2fv(tri[i].tc[2].data);
      glVertex3fv(vert[he->next->next->vert].data);
      glEnd();
    }
  }
}

void DefMesh::draw(int drawtex,int useColors)
{
  printDebug("%d\n",nvert,ntri);
  //glDisable(GL_LIGHTING);
  if(drawtex){
    glColor3f(1.0,1.0,1.0);
    glEnable(GL_TEXTURE_2D);
    //glBindTexture(GL_TEXTURE_2D,getTexture(useTexNumber));
    //glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
    //glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  }
  printErr("before setup");

  static int cnt=0;
  cnt++;
  if((cnt%10)==-1){
    printf("cnt %d\n",cnt);
    for(int i=0;i<ntri;i++)
    {
      HalfEdge * he=tri[i].edge;
      if(!he)continue;
      if(he->opp==0)printDebug("no opp %d\n",i);
      
      if(!drawtex){
	float c=(float)tri[i].flags/(float)lastnconn;
	glColor3f(c,1.0-c,c);
      }
      glBegin(GL_TRIANGLES);
      Vec3f temp=norm[he->vert]*normalDerivativeScale*sigma
	+vert[he->vert];
      glNormal3fv(norm[he->vert].data);
      glTexCoord2fv(tri[i].tc[0].data);
      glVertex3fv(temp.data);
      he=he->next;
      
      temp=norm[he->vert]*normalDerivativeScale*sigma
	+vert[he->vert];
      glNormal3fv(norm[he->vert].data);
      glTexCoord2fv(tri[i].tc[0].data);
      glVertex3fv(temp.data);
      he=he->next;

      temp=norm[he->vert]*normalDerivativeScale*sigma
	+vert[he->vert];
      glNormal3fv(norm[he->vert].data);
      glTexCoord2fv(tri[i].tc[0].data);
      glVertex3fv(temp.data);
      he=he->next;
      
      glEnd();
    }
  }
  else{
    for(int i=0;i<ntri;i++)
    {
      HalfEdge * he=tri[i].edge;
      if(!he)continue;
      if(he->opp==0)printDebug("no opp %d\n",i);
      
      if(!drawtex){
	float c=(float)tri[i].flags/(float)lastnconn;
	glColor3f(c,1.0-c,c);
      }
      glBegin(GL_TRIANGLES);
      glNormal3fv(norm[he->vert].data);
      glTexCoord2fv(tri[i].tc[0].data);
      if(useColors)
	glColor3fv(tri[i].col[0].data);
      glVertex3fv(vert[he->vert].data);
      glNormal3fv(norm[he->next->vert].data);
      glTexCoord2fv(tri[i].tc[1].data);
      if(useColors)
	glColor3fv(tri[i].col[1].data);
      glVertex3fv(vert[he->next->vert].data);
      glNormal3fv(norm[he->next->next->vert].data);
      glTexCoord2fv(tri[i].tc[2].data);
      if(useColors)
	glColor3fv(tri[i].col[2].data);
      glVertex3fv(vert[he->next->next->vert].data);
      glEnd();
    }
  }
}

/** Get the vertex neighbors of he->vert and return them in neighbors.
 \param he the half edge to use to find the neighbors of he->vert
 \param neighbors an array of pointer to HalfEdge
 \param max the number of elements in neighbors
*/ 
int vertexNeighbors(HalfEdge * he,HalfEdge ** neighbors,int max)
{
  int i=0;
  int nadj=0;
  HalfEdge * finger=he;
  while(i<max){
    finger=finger->next;
    if(nadj>=max)return max;
    neighbors[nadj++]=finger;
    finger=finger->next->opp;
    if(finger==he)break;
    i++;
  }
  return nadj;
}

/** Perform an axial transform (lechaud) on the mesh
 \param he1,he2 are valid halfEdges
 \pre he1 and he2 are about to collide, thus calls
 for the axial transform
 \post virtual neighbors are created around he1 and he2,
  and then face are added between the corresponding virtual neighbors 
  and the faces made with he1(resp. he2) and its virtual neighbors
  are removed.  Much like creating a tube between the two points
*/
void DefMesh::axialTform(HalfEdge * he1,HalfEdge * he2)
{
  HalfEdge * he1neigh[100];
  HalfEdge * he2neigh[100];
  int nhe1neigh,nhe2neigh;

  if(norm[he1->vert].dot(norm[he2->vert])>=0.8)return;

  makeVirtualNeighbors(he1,1);
  makeVirtualNeighbors(he2,1);
  
  //Find the two that are closest together
  nhe1neigh=vertexNeighbors(he1,he1neigh,100);
  nhe2neigh=vertexNeighbors(he2,he2neigh,100);
  
  if(nhe1neigh<=2 || nhe2neigh<=2){
    printDebug("warning: a single triangle floating around\n");
    //dosplit=0;
    he1=he2=0;
    return;
  }
  //Chances are these two are not colliding
  
  if(nhe1neigh>nhe2neigh){
    HalfEdge * backhe=he1;
    he1=he2;
    he2=backhe;
    //FIXME: this doesn't need to be recompouted
    nhe1neigh=vertexNeighbors(he1,he1neigh,100);
    nhe2neigh=vertexNeighbors(he2,he2neigh,100);
  }

  int nhe1i=0;
  int nhe2i=0;
  printDebug("getting closest\n");
  double closest=10e10;
  for(int i=0;i<nhe1neigh;i++){
    for(int j=0;j<nhe2neigh;j++){
      double dist=(vert[he1neigh[i]->vert]-vert[he2neigh[j]->vert]).len();
      if(dist<closest){
	closest=dist;
	nhe1i=i;
	nhe2i=j;
      }
    }
  }
  
  //triangulate starting from the closest
  //first fit a plane to each of the neighboring points
  Matrix m1(nhe1neigh,3);
  Matrix m2(nhe2neigh,3);
  Vec3f m1mean(0,0,0),m2mean(0,0,0);
  for(int i=0;i<nhe1neigh;i++){
    m1(i,0)=vert[he1neigh[i]->vert].x;
    m1(i,1)=vert[he1neigh[i]->vert].y;
    m1(i,2)=vert[he1neigh[i]->vert].z;
    m1mean+=vert[he1neigh[i]->vert];
  }
  for(int i=0;i<nhe2neigh;i++){
    m2(i,0)=vert[he2neigh[i]->vert].x;
    m2(i,1)=vert[he2neigh[i]->vert].y;
    m2(i,2)=vert[he2neigh[i]->vert].z;
    m2mean+=vert[he2neigh[i]->vert];
  }
  m1mean*=(1.0/nhe1neigh);
  m2mean*=(1.0/nhe2neigh);
  
  //subtract mean
  for(int i=0;i<nhe1neigh;i++){
    m1(i,0)-=m1mean.x;m1(i,1)-=m1mean.y;m1(i,2)-=m1mean.z;
  }
  for(int i=0;i<nhe2neigh;i++){
    m2(i,0)-=m2mean.x;m2(i,1)-=m2mean.y;m2(i,2)-=m2mean.z;
  }
  //compute the rotation
  Matrix u,s,v1,v2;
  //m1.svd(u,s,v1);
  //m2.svd(u,s,v2);
  v1=Matrix::eye(3,3);
  v2=Matrix::eye(3,3);
  Matrix p1=m1*v1;
  Matrix p2=m2*v2;
  double ne1angles[100];//stop being so lazy
  double ne2angles[100];
  
  for(int i=0;i<nhe1neigh;i++){
    ne1angles[i]=atan2(p1(i,1),p1(i,0));
    while(ne1angles[i]<0)ne1angles[i]+=2*M_PI;
    printDebug("%d %f\n",0,ne1angles[i]);
  }
  for(int i=0;i<nhe2neigh;i++){
    ne2angles[i]=atan2(p2(i,1),p2(i,0));
    while(ne2angles[i]<0)ne2angles[i]+=2*M_PI;
    printDebug("%d %f\n",1,ne2angles[i]);
  }
  
  //make sure that angles are going in the positive direction
  //otherwise one of the axis in the rotation matrix may be wrong....
  //printDebug("determinants: %f %f\n",v1.det(),v2.det());
  int wrap=0;
  double last=ne1angles[nhe1neigh-1];
  for(int i=0;i<nhe1neigh;i++){
    if(ne1angles[i]<last)wrap++;
    last=ne1angles[i];
  }
  if(wrap>=2){
    for(int i=0;i<nhe1neigh;i++)
      ne1angles[i]*=-1;
  }
  wrap=0;
  last=ne2angles[nhe2neigh-1];
  for(int i=0;i<nhe2neigh;i++){
    if(ne2angles[i]<last)wrap++;
    last=ne2angles[i];
  }
  if(wrap>=2){
    for(int i=0;i<nhe2neigh;i++)
      ne2angles[i]*=-1;
  }
  
  //recompute the angles from the nearest
  printDebug("angles1:");
  double nhe1iangle=ne1angles[nhe1i];
  for(int i=0;i<nhe1neigh;i++){
    ne1angles[i]-=nhe1iangle;
    while(ne1angles[i]<0)ne1angles[i]+=2*M_PI;
    printDebug("%f ",ne1angles[i]);
  }printDebug("\nangles2:");
  double nhe2iangle=ne2angles[nhe2i];
  for(int i=0;i<nhe2neigh;i++){
    ne2angles[i]-=nhe2iangle;
    ne2angles[i]*=-1;
    while(ne2angles[i]<0)ne2angles[i]+=2*M_PI;
    printDebug("%f ",ne2angles[i]);
  }printDebug("\n");
  
  
  /*for(int i=1;i<nhe1neigh;i++){
    int cur=(i+nhe1i)%nhe1neigh;
    int prev=(i-1+nhe1i+nhe1neigh)%nhe1neigh;
    if(ne1angles[cur]<ne1angles[prev])*/
  {
    int cur=0;
    double ang=0;
    double anginc=2.0*M_PI/nhe1neigh;
    for(int k=0;k<nhe1neigh;k++){
      cur=(k+nhe1i)%nhe1neigh;
      ne1angles[cur]=ang;
      ang+=anginc;
    }
    //printDebug("!!!!!!!!!!!!!!!!!!!!!!enforcing increasing\n");
    //break;
  }
  /*
    for(int i=2;i<nhe2neigh;i++){
    int cur=(i+nhe2i)%nhe2neigh;
    int prev=(i-1+nhe2i+nhe2neigh)%nhe2neigh;
    if(ne2angles[cur]>ne2angles[prev])
  */
  {
    int cur=0;
    double ang=0;
    double anginc=2.0*M_PI/nhe2neigh;
    ne2angles[0]=0;
    for(int k=1;k<nhe2neigh;k++){
      cur=(k+nhe2i)%nhe2neigh;
      ne2angles[cur]=2*M_PI-ang;
      ang+=anginc;
    }
    //printDebug("!!!!!!!!!!!!!!!!!!!!!!enforcing increasing\n");
    //break;
  }
  //}
  
  printDebug("#neighors(he1)=%d, [",nhe1neigh);
  for(int i=0;i<nhe1neigh;i++)printDebug("%d ",he1neigh[i]->vert);
  printDebug("]\n");
  
  printDebug("#neighors(he2)=%d, [",nhe2neigh);
  for(int i=0;i<nhe2neigh;i++)printDebug("%d ",he2neigh[i]->vert);
  printDebug("]\n");
  
  //Create the triangles....
  int base=ntri;
  int j=0;
  int created=0;
  int prevhe1i=-1;
  //<= so that the first vertex has a chance to pick up anything
  //that didn't get triangulated
  for(int i=0;i<=nhe1neigh;i++){
    int nexthe1i=(nhe1i+1)%nhe1neigh;
    for(;j<=nhe2neigh;j++){
      double diff1=fabs(ne1angles[nexthe1i]-ne2angles[nhe2i]);
      double diff2=fabs(ne1angles[nhe1i]-ne2angles[nhe2i]);
      if(diff1>=2*M_PI)diff1-=2*M_PI;
      if(diff2>=2*M_PI)diff2-=2*M_PI;
      int nexthe2i=(nhe2i-1+nhe2neigh)%nhe2neigh;
      //join the prevhe1i and nexthe1i
      if(diff1<diff2){
	//if(ne2angles[nhe2i]>ne1angles[nexthe1i]){
	//if(prevhe1i!=-1)
	{
	  printDebug("%d,%f gets %d,%f\n",nhe1i,ne1angles[nhe1i],nhe2i
		 ,ne2angles[nhe2i]);
	  
	  //letting a=nhe1i b=nexthe1i c=nexthe2i
	  //need to create triangle abc
	  HalfEdge * ab, * bc, * ca;
	  allocFace(&ab,&bc,&ca);
	  //modified->push_back(ab->face);
	  
	  int a=he1neigh[nhe1i]->vert;
	  int b=he1neigh[nexthe1i]->vert;
	  int c=he2neigh[nhe2i]->vert;//nexthe2i
	  ab->vert=a;
	  bc->vert=b;
	  ca->vert=c;
	  
	  ab->opp=he1neigh[nhe1i]->opp;
	  ab->opp->opp=ab;
	  
	  printDebug("creating abc: %d %d %d\n",a,b,c);
	  printDebug("ab=%d-%d opp %d-%d\n",ab->vert,ab->next->vert,
		 ab->opp->vert,ab->opp->next->vert);
	  //still need to do bc->opp and ca->opp
	  created++;//created a new faces
	}
	break;
      }
      else{
	
	if(i==nhe1neigh && j==nhe2neigh)break;
	//need to create triangle acd
	HalfEdge * ac, * cd, * da;
	allocFace(&ac,&cd,&da);
	//modified->push_back(ac->face);

	int a=he1neigh[nhe1i]->vert;
	int c=he2neigh[nexthe2i]->vert;
	int d=he2neigh[nhe2i]->vert;
	
	ac->vert=a;
	cd->vert=c;
	da->vert=d;
	
	cd->opp=he2neigh[nexthe2i]->opp;
	cd->opp->opp=cd;
	printDebug("creating acd: %d %d %d\n",a,c,d);
	printDebug("cd=%d-%d opp %d-%d\n",cd->vert,cd->next->vert,
	       cd->opp->vert,cd->opp->next->vert);
	
	//still need to do ac->opp and da->opp
	created++;
      }
      nhe2i=nexthe2i;
    }
    prevhe1i=nhe1i;
    nhe1i++;
    if(nhe1i>=nhe1neigh)nhe1i=0;
  }
  printDebug("created %d faces, total neighbors %d\n",created,nhe1neigh+nhe2neigh);
  
  
  for(int i=0;i<created;i++){
    HalfEdge * h=tri[i+base].edge;
    for(int j=0;j<3;j++,h=h->next){
      if(h->opp==0){
	HalfEdge * nextTri=tri[(i+1)%created+base].edge;
	for(int k=0;k<3;k++,nextTri=nextTri->next){
	  if(h->next->vert==
	     nextTri->vert &&
	     nextTri->next->vert==
	     h->vert){
	    h->opp=nextTri;
	    h->opp->opp=h;
	    break;
	  }
	}
	//if(h->opp)continue;
	nextTri=tri[(i-1+created)%created+base].edge;
	for(int k=0;k<3;k++,nextTri=nextTri->next){
	  if(h->next->vert==
	     nextTri->vert &&
	     nextTri->next->vert==
	     h->vert){
	    h->opp=nextTri;
	    h->opp->opp=h;
	    break;
	  }
	}
	
	if(h->opp==0){
	  printDebug("%d %d didn't find an opp\n",h->vert,h->next->vert);
	}
	else
	  printDebug("%d-%d found opp\n",h->vert,h->next->vert);
      }
    }
  }
  printDebug("deleting\n");
  //Now delete all adjacent triangles to the vertices, and 
  //the vertices themselves
  for(int i=0;i<nhe1neigh;i++){
    HalfEdge * he=he1neigh[i];
    if(he->opp==he){
      printDebug("deleting something still pointing to it\n");
    }
    tri[he->face].edge=0;
    he->next->next->opp=0;
    he->next->next->next=0;
    po.removeUserData(vert[he->next->next->vert]);
    delete he->next->next;
    he->next->opp=0;
    he->next->next=0;
    po.removeUserData(vert[he->next->vert]);
    delete he->next;
    he->opp=0;
    he->next=0;
    po.removeUserData(vert[he->vert]);
    delete he;
  }
  for(int i=0;i<nhe2neigh;i++){
    HalfEdge * he=he2neigh[i];
    if(he->opp==he){
      printDebug("deleting something still pointing to it\n");
    }
    tri[he->face].edge=0;
    he->next->next->opp=0;
    he->next->next->next=0;
    po.removeUserData(vert[he->next->next->vert]);
    delete he->next->next;
    he->next->opp=0;
    he->next->next=0;
    po.removeUserData(vert[he->next->vert]);
    delete he->next;
    he->opp=0;
    he->next=0;
    po.removeUserData(vert[he->vert]);
    delete he;
  }
  printDebug("checking conssistency\n");
  checkConsistency("After tube\n");
}


/** The topological neighbors of he are fixed to have the same vertex.
    All the neighbors of the HalfEdge he that should have the same
    vert pointer, are updated so that they actually do.
    \pre he is a valid HalfEdge pointer
    \post The vert pointers of the neighbors are fixed.
*/
void DefMesh::fixNeighboringVert(HalfEdge * he)
{
  int i=0;
  int max=100;
  int newvert=he->vert;
  HalfEdge * finger=he->opp->next;
 
  while(finger!=he && i<max){
    finger->vert=newvert;
    finger=finger->opp->next;
    i++;
  }
  if(i==max){
    printDebug("there is a good chance fixNeighboringVert did not finish\n");
    exit(0);
  }
}

/** The topological neighbors of the face are fixed to have the same vertex.
    \pre face is a valid index to the faces.
    \post all neighboring HalfEdges have their vert pointer fixed.
    \sa DefMesh::fixNeighboringVert(HalfEdge * he)
*/
void DefMesh::fixNeighboringVert(int face)
{
  HalfEdge * he=tri[face].edge;
  if(he==0){
    printDebug("fixNeighboringVert (%s,%d), tri[%d]==0\n",
	   __FILE__,__LINE__,face);
    exit(0);
  }
  fixNeighboringVert(he);
  fixNeighboringVert(he->next);
  fixNeighboringVert(he->next->next);
}

/** A consistency check on the topology of the mesh
    This is only valid if CHECK is defined.  It is
    very slow, and used only for debug purposes.
    \pre str is an error string to print.
    \post any errors are displayed, and the program exits if
    there are any errors.
*/
void DefMesh::checkConsistency(char * str)
{
#ifndef CHECK
  return;
#endif
  int nerr=0;
  for(int i=0;i<ntri;i++){
    HalfEdge * he=tri[i].edge;
    if(!he)continue;
    for(int j=0;j<3;j++){
      if(he->opp->opp!=he){
	printDebug("err: opposite is not set correctly %d %d\n",i,j);
	printDebug("     (%d,%p):%d-%d opposite: (%d,%p):%d-%d\n",
	       he->face,he,he->vert,he->next->vert,
	       he->opp->face,he->opp,he->opp->vert,he->opp->next->vert);
	nerr++;
      }
      if(he->opp->next->vert!=he->vert){
	printDebug("err: opposite's shouldn't be opposite %d %d\n",i,j);
	printDebug("     (%d,%p):%d-%d opposite: (%d,%p):%d-%d\n",
	       he->face,he,he->vert,he->next->vert,
	       he->opp->face,he->opp,he->opp->vert,he->opp->next->vert);
	nerr++;
      }
      if(he->face!=i){
	printDebug("err: face is not set correctly\n");
	nerr++;
      }
      if(he->next->next->next!=he){
	printDebug("err: inner triangle not set correctly\n");
	nerr++;
      }
      if(he->next==he){
	printDebug("err: he is itself\n");
	nerr++;
      }
      if(he->opp==he){
	printDebug("err: he opp is itself\n");
	nerr++;
      }
      if(!tri[he->opp->face].edge){
	printDebug("err: opp=%d of %d has been deleted\n",he->opp->face,i);
	nerr++;
      }
      if(he->vert==he->next->vert){
	printDebug("err: degenerate edge %d %d\n",he->vert,he->next->vert);
	//nerr++;
      }
      HalfEdge * finger = he;
      int v=he->vert;
      int attempts=0;
      while(attempts<100){
	if(finger->vert!=v){
	  printDebug("err: adjacent vertices are screwed up for vert %d (%d,%d)\n",
		 he->vert,i,j);
	  nerr++;
	  break;
	}
	finger=finger->opp->next;
	if(finger==he)break;
	attempts++;
      }
      if(attempts==100){
	printDebug("err: unable to finish checking neighbors\n");
	nerr++;
      }
      he=he->next;
    }
  } 
  if(nerr){
    printDebug("nerrs %d at %s\n",nerr,str);
    exit(0);
  }
}

/** Generate a layer of neighbors between the given half edge
  New vertices are created on all edges from he->vert.  The
  corresponding faces are also created.
  \param he the half edge whose vert to create neighbors about
  \param removeFromOctree if this is set, the vertex will be removed
  from the eoctree
  \pre he is not null
  \post the virtual neighbors are created
*/
void DefMesh::makeVirtualNeighbors(HalfEdge * he,int removeFromOctree)
{
  HalfEdge * neigh[100];
  int newneigh[100];
  int nneigh=0;
  HalfEdge * finger=he;
  Vec3f v=vert[he->vert];
  Vec3f n=norm[he->vert];
  printDebug("making virtual neighbors for %p,%d\n",he,he->vert);
  do{
    //Add all affected faces to the modified list
    //modified->push_back(finger->next->opp->face);
    //modified->push_back(finger->face);

    neigh[nneigh]=finger->next;
    if(removeFromOctree)
      po.removeUserData(vert[neigh[nneigh]->vert]);
    Vec3f newv=(v+vert[neigh[nneigh]->vert])*0.5;
    Vec3f newnorm=(n+norm[neigh[nneigh]->vert])*0.5;
    ensureVert(nvert+1);
    newneigh[nneigh++]=nvert;
    vert[nvert]=newv;
    norm[nvert]=newnorm;
    nvert++;
    finger=finger->next->next->opp;
  }while(finger!=he);
  printDebug("it has %d neighbors, base is %d, neighbors : ",nneigh,ntri);
  for(int i=0;i<nneigh;i++)printDebug("%d ",neigh[i]->vert);
  printDebug("\n");

  
  //assuming vab goes to vcd abc cbd
  // v__b
  // | /
  // a 
  //
  // v__d___b
  // | / / /
  // |//  /
  // c  _/
  // | /
  // a/
  //where a2 b2 are the new verts
  int base=ntri;
  for(int i=0;i<nneigh;i++){
    HalfEdge * ab,* bc,* ca;
    HalfEdge * cb,* bd,* dc;
    HalfEdge * vc,*cd,*dv;
    int c=newneigh[i];
    int d=newneigh[(i+1)%nneigh];
    int a=neigh[i]->vert;
    int b=neigh[(i+1)%nneigh]->vert;
    cd=neigh[i];
    dv=cd->next;
    vc=dv->next;

    allocFace(&ab,&bc,&ca);
    allocFace(&cb,&bd,&dc);
    //modified->push_back(ab->face);
    //modified->push_back(cb->face);

    ab->vert=a;
    bc->vert=b;
    ca->vert=c;

    cb->vert=c;
    bd->vert=b;
    dc->vert=d;

    ab->opp=cd->opp;
    ab->opp->opp=ab;
    
    cd->opp=dc;
    dc->opp=cd;

    bc->opp=cb;
    cb->opp=bc;

    //set ca->opp and bd->opp
  }
  //not sure if this can be done in the above loop, it may
  //overwrite a and b that are needed
  for(int i=0;i<nneigh;i++){
    int c=newneigh[i];
    int d=newneigh[(i+1)%nneigh];
    HalfEdge * vc,*cd,*dv;
    cd=neigh[i];
    dv=cd->next;
    vc=dv->next;

    cd->vert=c;
    dv->vert=d;
    vc->vert=he->vert;
  }
  printDebug("fixing up adj\n");
  //Fix up the adjacencies..this should be possible in a
  //couple less iterations
  for(int i=base;i<ntri;i+=2){
    int face1=i;
    int face1adj=(face1+2*nneigh-1-base)%(2*nneigh)+base;
    int face2=i+1;
    int face2adj=(face2+1-base)%(2*nneigh)+base;
    HalfEdge * ca=tri[face1].edge->next->next;
    HalfEdge * bd=tri[face2].edge->next;

    HalfEdge * ca_opp=tri[face1adj].edge->next; //bd of the adj face
    HalfEdge * bd_opp=tri[face2adj].edge->next->next; //ca of the adj face

    ca->opp=ca_opp;
    ca->opp->opp=ca;
    bd->opp=bd_opp;
    bd->opp->opp=bd;
  }
  printDebug("checking consistency\n");
  checkConsistency("After making virtual neighbors\n");

}

/** remove any unused triangles
 \post removes the triangles that are not used
 */
void DefMesh::removeUnusedTri()
{
  int to=0;
  int from=0;
  for(int from=0;from<ntri;from++){
    //pointers are the same, data exists, just continue..
    //only happens at the beginning
    if(to==from && tri[from].edge)to++;
    else if(tri[from].edge){
      tri[from].edge->next->next->face=to;
      tri[from].edge->next->face=to;
      tri[from].edge->face=to;
      tri[to++]=tri[from];
      tri[from].edge=0;
    }
  }
  printDebug("%d/%d\n",to,ntri);
  ntri=to;
}

/** Distance handler for potential collisions of he1 and he2
 */
int dh(DefMesh * dm,HalfEdge * he1,HalfEdge * he2)
{
  if(!he1 || !he2)return 0;
  if(!he1->opp) return 0;
  if(!he1->next->opp) return 0;
  if(!he1->next->next->opp) return 0;
  if(!he2->opp) return 0;
  if(!he2->next->opp) return 0;
  if(!he2->next->next->opp) return 0;

  HalfEdge * neighbors[100];
  int nneigh,i;
  
  if(he1->vert==he2->vert)return 0;
  /*
    printf("nodePoint %f %f %f\n",nodePoint.x,nodePoint.y,nodePoint.z);
    printf("vert[%f %f %f]\n",dm.vert[he1->vert].x,
    dm.vert[he1->vert].y,dm.vert[he1->vert].z);
  */
  
  nneigh=vertexNeighbors(he1,neighbors,100);
  for(i=0;i<nneigh;i++){
    if(he2->vert==neighbors[i]->vert)break;
  }
  if(i==nneigh){
    //printf("doing axial tform\n");
    dm->po.removeUserData(dm->vert[he1->vert]);
    dm->po.removeUserData(dm->vert[he2->vert]);
    dm->axialTform(he1,he2);
    //printf("done dh\n");
    return 1;
    //printf("found one\n");
    //dm.tri[he1->face].flags=4;
    //dm.tri[he2->face].flags=4;
  }
  return 0;
}

/** Distance handler for potential collisions of he1 and he2
 */
int dh(const Vec3f & nodePoint,void * nodeData,
	const Vec3f & queryPoint,void * queryData,
	double rad)
{
  return dh(0,(HalfEdge *)nodeData,(HalfEdge *)queryData);
}

void DefMesh::lechaudRemeshHelper(int face)
{
  HalfEdge * he=tri[face].edge;
  if(!he)return;
  if(!he->opp){
    return;
  }
  if(!he->next->opp){
    return;
  }
  if(!he->next->next->opp){
    return;
  }
  Vec3f a;
  Vec3f b;
    
  int degenerate=0;
  for(int j=0;j<3;j++){
    HalfEdge * prev=he;
    he=he->next;
    if(prev->vert==he->vert){
      int face=he->face;
      he->opp->opp=he->next->opp;
      he->next->opp->opp=he->opp;
      
      HalfEdge * other=prev->opp;
      if(other->face!=face){
	printDebug("need to take care of this...\n");
	printDebug("other->face(%d) %d %d %d\n",other->face,
		   other->vert,other->next->vert,other->next->next->vert);
	HalfEdge * eq=0;
	if(other->vert==other->next->vert)
	  eq=other->next;
	if(other->next->vert==other->next->next->vert)
	  eq=other->next->next;
	if(eq){
	  eq->opp->opp=eq->next->opp;
	  eq->next->opp->opp=eq->opp;
	  tri[eq->face].edge=0;
	  delete eq->next->next;
	  delete eq->next;
	  delete eq;
	  
	}  
      }
      delete prev;
      delete he->next;
      delete he;
      tri[face].edge=0;
      
      //getchar();
      degenerate=1;
      break;
    }
  }
  if(degenerate)return;
  
  he=tri[face].edge;
  for(int j=0;j<3;j++){
    HalfEdge * prev=he;
    a=vert[prev->vert];
    he=he->next;
    b=vert[he->vert];
    
    Vec3f edge=b-a;
    double edgelen=edge.len();
    
    if(edgelen<sigma){
      //perform annular here
      HalfEdge * prevAdj[100];
      int isectvert[100];
      int nprevadj=0;
      HalfEdge * heAdj[100];
      int nheadj=0;
      int nisect=0;
      HalfEdge * finger;
      HalfEdge * heMatch=0;
      HalfEdge * prevMatch=0;
      nprevadj=vertexNeighbors(prev,prevAdj,100);
      nheadj=vertexNeighbors(he,heAdj,100);
      
      for(int k=0;k<nheadj;k++){
	for(int j=0;j<nprevadj;j++){
	  if(heAdj[k]->vert==prevAdj[j]->vert){
	    isectvert[nisect]=heAdj[k]->vert;
	    nisect++;
	    if(nisect>=3 && heMatch==0){
	      if(heAdj[k]->next->vert != prev->vert){
		heMatch=heAdj[k];
		prevMatch=prevAdj[j];
	      }
	    }
	  }
	}
      }
      if(nisect>=3){
	if(heMatch==0){
	  printDebug("there is no other vertex that is a neighbor of both but not a face of the mesh, what to do? nisect==%d\n",nisect);
	  //exit(0);
	}
	else{
	  //need to duplicate u==prev->vert
	  //                  v==he->vert
	  //             and  o==heMatch->vert==prevMatch->vert
	  //need to split along uv vo ou
	  //uv is given by prev and prev->opp
	  //vo is given by heMatch->next->next and the opp
	  //ou is given by prevMatch->next->next->opp and the opp
	  
	  printDebug("nisect (adj: %d,%d, isect: %d)\n",nprevadj,nheadj,
		   nisect);
	  
	  printDebug("prevadj(%d): ",prev->vert);
	  for(int h=0;h<nprevadj;h++){
	    printDebug("%d ",prevAdj[h]->vert);
	  }
	  printDebug("\n");
	  printDebug("nheadj(%d): ",he->vert);
	  for(int h=0;h<nheadj;h++){
	    printDebug("%d ",heAdj[h]->vert);
	  }
	  printDebug("\n");
	  printDebug("isect: ");
	  for(int h=0;h<nisect;h++){
	    printDebug("%d ",isectvert[h]);
	  }
	  printDebug("\n");

	  printDebug("melting\n");
	  ensureVert(nvert+3);
	  int u2,v2,o2;
	  int u=prev->vert,v=he->vert,o=heMatch->vert;
	  HalfEdge * v2u2,* u2o2,* o2v2;
	  HalfEdge * u2v2=prev;
	  HalfEdge * v2o2=heMatch->next->next;
	  HalfEdge * o2u2=prevMatch->next->next->opp;
	  HalfEdge * u1v1,* v1o1,* o1u1;
	  HalfEdge * v1u1=u2v2->opp;
	  HalfEdge * o1v1=v2o2->opp;
	  HalfEdge * u1o1=o2u2->opp;

	  allocFace(&v2u2,&u2o2,&o2v2);
	  allocFace(&u1v1,&v1o1,&o1u1);
	    
	  u2=nvert;
	  v2=nvert+1;
	  o2=nvert+2;
	  nvert+=3;

	  vert[u2]=vert[prev->vert];
	  vert[v2]=vert[he->vert];
	  vert[o2]=vert[heMatch->vert];
#warning "fix normals after split"

	  u2v2->vert=u2;
	  v2o2->vert=v2;
	  o2u2->vert=o2;

	  v2u2->vert=v2;
	  u2o2->vert=u2;
	  o2v2->vert=o2;
	    
	  u2v2->opp=v2u2;
	  v2u2->opp=u2v2;	    
	  v2o2->opp=o2v2;
	  o2v2->opp=v2o2;
	  o2u2->opp=u2o2;
	  u2o2->opp=o2u2;

	  u1v1->vert=u;
	  v1o1->vert=v;
	  o1u1->vert=o;
  
	  v1u1->opp=u1v1;
	  u1v1->opp=v1u1;
	  v1o1->opp=o1v1;
	  o1v1->opp=v1o1;
	  u1o1->opp=o1u1;
	  o1u1->opp=u1o1;

	  fixNeighboringVert(v2u2->face);
	    
	  //tri[v2u2->face].flags=1;
	  //tri[u1v1->face].flags=1;

	  printDebug("u=%d->%d %d\n",u,u,u2);
	  printDebug("v=%d->%d %d\n",v,v,v2);
	  printDebug("o=%d->%d %d\n",o,o,o2);
	    
	  int nu=0;
	  int nv=0;

	  finger=u1v1;;
	  printDebug("u1: ");
	  do{
	    printDebug("%d ",finger->next->vert);
	    finger=finger->opp->next;
	    nu++;
	  }while(finger!=u1v1);

	  printDebug("\n");
	  finger=v1o1;
	  printDebug("v1: [");
	  do{
	    printDebug("%d ",finger->next->vert);
	    finger=finger->opp->next;
	    nv++;
	  }while(finger!=v1o1);
	  printDebug("]\n");

	  finger=u2v2;
	  printDebug("u2: [");
	  do{
	    printDebug("%d ",finger->next->vert);
	    finger=finger->opp->next;
	    nu++;
	  }while(finger!=u2v2);

	  printDebug("]\n");
	  finger=v2o2;
	  printDebug("v2: [");
	  do{
	    printDebug("%d ",finger->next->vert);
	    finger=finger->opp->next;
	    nv++;
	  }while(finger!=v2o2);
	  printDebug("]\n");

	  checkConsistency("bottom of melting");
	  printDebug("%d %d %d\n",nprevadj,nheadj,nisect);
	  if(nu!=(nprevadj+2) ||
	     nv!=(nheadj+2)){
	    printf("inconsistent nu=%d!=%d nv=%d!=%d\n",nu,nprevadj+2,
		   nv,nheadj+2);
	    exit(0);
	  }
	}
      }
      else {
	checkConsistency("top of removing");
	Vec3f newVert=(a+b)*0.5;
	vert[prev->vert]=newVert;
	vert[he->vert]=newVert;
#warning "FIXME: lingering vert"
	int hevertb4=he->vert;
	finger=he;
	//printDebug("working on triangle %d, removing %d %d\n",i,he->face,prev->opp->face);
	if(finger->vert==prev->vert){
	  printDebug("edge has already been collapsed...pointer not fixed?\n");
	}
	  
	while(finger!=prev->opp){
	  if(finger->vert!=hevertb4){
	    printDebug("check for vert failed in delete %d %d\n",finger->vert,hevertb4);
	    exit(0);
	  }
	  //else
	  // printDebug("changed %d %d %p %p\n",finger->vert,prev->vert,finger,finger->opp->next);
	  finger->vert=prev->vert;
	  finger=finger->opp->next;
	}

	he->opp->opp=he->next->opp;
	he->next->opp->opp=he->opp;
	  
	prev->opp->next->opp->opp=prev->opp->next->next->opp;
	prev->opp->next->next->opp->opp=prev->opp->next->opp;
	  
#warning "FIXME: lingering triangles" 
	tri[prev->opp->face].edge=0;
	tri[prev->face].edge=0;

	delete prev->opp->next->next;
	delete prev->opp->next;
	delete prev->opp;

	delete he->next->next;
	delete he->next;
	delete he;
	checkConsistency("bottom of removing");
	break;
      }
    }

    //"creation or inversion"
    else if(edgelen>=2.5*sigma){
      //printDebug("creating %d %f %d %d %d\n",i,edgelen,he->vert,prev->vert,nvert);
      //printDebug("%f %f %f -- %f %f %f\n",a.x,a.y,a.z,b.x,b.y,b.z);
	

      Vec3f newVert=(a+b)*0.5;
      if((newVert-vert[he->next->vert]).len()>edgelen ||
	 (newVert-vert[prev->opp->next->next->vert]).len()>edgelen){
	 
      }
      else{
	checkConsistency("top of insert");
	Vec3f newNorm=(norm[prev->vert]+norm[he->vert])*0.5;
	newNorm.normalize();
	ensureVert(nvert+1);
	vert[nvert]=newVert;
	norm[nvert]=newNorm;
	int e=nvert;
	int a=he->vert;
	int b=he->next->vert;
	int c=prev->vert;
	int d=prev->opp->next->next->vert;

	if(b==d || a==c){
	  printDebug("small connected component...ignoring\n");
	  //tri[he->face].flags=2;
	  //tri[prev->opp->face].flags=2;
	}
	else{
	  HalfEdge * abopp=he->opp;
	  HalfEdge * daopp=prev->opp->next->next->opp;
	  nvert++;
	    
	  //Replace triangle abc acd with
	  // ebc,ecd,bea,eda where
	  // e is the new vertex
	    
	  printDebug("verts: %d %d %d %d faces %d %d adjacent faces %d %d\n",
		     a,b,c,d,prev->opp->face,prev->face,abopp->face,daopp->face);
	  printDebug("inserting %d between %d %d two faces\n",nvert,prev->vert,he->vert);
	    
	    
	  //First Triangle, opposite to he..
	  HalfEdge * be,* ea,* ab;
	  HalfEdge * ed,* da,* ae;
	  allocFace(&be,&ea,&ab);
	  allocFace(&ed,&da,&ae);
	    
	  printDebug("creating faces %d %d\n",be->face,ed->face);
	    
	  he->vert=e;
	  prev->opp->vert=e;
	    
	  be->vert=b;
	  ea->vert=e;
	  ab->vert=a;
	    
	  ed->vert=e;
	  da->vert=d;
	  ae->vert=a;
	    
	  be->opp=he;
	  he->opp=be;
	    
	  ea->opp=ae;
	  ae->opp=ea;
	    
	  ab->opp=abopp;
	  if(abopp)
	    abopp->opp=ab;
	    
	  ed->opp=prev->opp->next->next;
	  ed->opp->opp=ed;
	    
	  da->opp=daopp;
	  if(daopp)
	    daopp->opp=da;
	}
	checkConsistency("Bottom of inserting\n");
	//return;
      }
    }
  }
}

/** Remesh the deformable structure using Lechauds method.
    The vertices should be evenly spaced, and any collisions
    or separations will be removed.
 */
int DefMesh::lechaudRemesh()
{
  //if(lastModified)delete lastModified;
  //lastModified=modified;
  //modified = new vector<int>();
  
  int oldntri=ntri;
  int i=0;
  for(i=0;i<ntri;i++){
    lechaudRemeshHelper(i);
  }
    //po.removeAll();
  //po.setDistanceHandler(dh);
  const int maxhits=1000;
  int nhits=0;
  void * hits[maxhits];
  po.resize((int)2*(pow(nvert,0.333)+2));
  
  Vec3f min( 10e10, 10e10, 10e10);
  Vec3f max(-10e10,-10e10,-10e10);
  for(int i=0;i<nvert;i++){
    min.x=MIN(min.x,vert[i].x);
    min.y=MIN(min.y,vert[i].y);
    min.z=MIN(min.z,vert[i].z);
    max.x=MAX(max.x,vert[i].x);
    max.y=MAX(max.y,vert[i].y);
    max.z=MAX(max.z,vert[i].z);
  }
  //printf("%f-%f %f-%f %f-%f\n",min.x,max.x,min.y,max.y,min.z,max.z);
  po.setRanges(min.x,max.x,min.y,max.y,min.z,max.z);
  checkConsistency("before distance stuff\n");
  printDebug("inserting into point octree\n");

  double dist=sigma*2.5/(sqrt(3));
  oldntri=ntri;
  for(int i=0;i<oldntri;i++){
    if(!tri[i].edge)continue;
    po.insert(vert[tri[i].edge->vert],tri[i].edge);
    po.insert(vert[tri[i].edge->next->vert],tri[i].edge->next);
    po.insert(vert[tri[i].edge->next->next->vert],tri[i].edge->next->next);

    if(!tri[i].edge)continue;
    nhits=po.distanceQuery(vert[tri[i].edge->vert],dist,hits,maxhits);
    for(int hiti=0;hiti<nhits;hiti++)
      if(dh(this,(HalfEdge *)hits[hiti],tri[i].edge))break;
    
    if(!tri[i].edge)continue;
    nhits=po.distanceQuery(vert[tri[i].edge->next->vert],dist,hits,maxhits);
    for(int hiti=0;hiti<nhits;hiti++)
      if(dh(this,(HalfEdge *)hits[hiti],tri[i].edge->next))break;

    if(!tri[i].edge)continue;
    nhits=po.distanceQuery(vert[tri[i].edge->next->next->vert],dist,hits,maxhits);
    for(int hiti=0;hiti<nhits;hiti++)
    if(dh(this,(HalfEdge *)hits[hiti],tri[i].edge->next->next))break;
  }
  /*
  printDebug("done inserting\n");
  double dist=sigma*2.5/(sqrt(3));
  oldntri=ntri;
  for(i=0;i<oldntri;i++){
    if(!tri[i].edge)continue;
    po.distanceQuery(vert[tri[i].edge->vert],dist,tri[i].edge);
    if(!tri[i].edge)continue;
    po.distanceQuery(vert[tri[i].edge->next->vert],dist,tri[i].edge->next);
    if(!tri[i].edge)continue;
    po.distanceQuery(vert[tri[i].edge->next->next->vert],dist,tri[i].edge->next->next);
    }*/
  removeUnusedTri();
  checkConsistency("end of lechaud");
  initNormals();
  printDebug("added %d\n",ntri-oldntri);
  
  /*printf("added %d/%d, modified %d\n",
	 ntri-oldntri,
	 ntri,modified->size());*/
  return ntri-oldntri;
}

void DefMesh::smooth(float param,vector <int> * onlydo)
{
  Vec3f * smoothedVert=new Vec3f[nvert];
  int * cnt=new int[nvert];
  memset(cnt,0,sizeof(int)*nvert);
  memset(smoothedVert,0,sizeof(Vec3f)*nvert);

  for(int i=0;i<ntri;i++){
    HalfEdge * he=tri[i].edge;
    if(!he)continue;
    for(int k=0;k<3;k++,he=he->next){
      if(!cnt[he->vert]){
	HalfEdge * finger=he;
	while(1){
	  finger=finger->next;
	  smoothedVert[he->vert]+=vert[finger->vert];
	  cnt[he->vert]++;
	  finger=finger->next->opp;
	  if(finger==he)break;
	}
      }
    }
  }
  if(onlydo==0){
    for(int i=0;i<nvert;i++){
      if(cnt[i]){
	smoothedVert[i]*=(1.0/(double)cnt[i]);
	vert[i]*=(1-param);
	vert[i]+=(smoothedVert[i]*param);
      }
    }
  }
  else{
    for(int i=0;i<onlydo->size();i++){
      int v=(*onlydo)[i];
      if(cnt[v]){
	smoothedVert[v]*=(1.0/(double)cnt[v]);
	vert[v]*=(1-param);
	vert[v]+=(smoothedVert[v]*param);
      }
    }
  }
  delete cnt;
  delete smoothedVert;
}

void DefMesh::ranges(Vec3f & min,Vec3f & max){
  min=Vec3f(10e10,10e10,10e10);
  max=Vec3f(-10e10,-10e10,-10e10);
  for(int i=0;i<nvert;i++){
    min.x=MIN(min.x,vert[i].x);
    min.y=MIN(min.y,vert[i].y);
    min.z=MIN(min.z,vert[i].z);
    max.x=MAX(max.x,vert[i].x);
    max.y=MAX(max.y,vert[i].y);
    max.z=MAX(max.z,vert[i].z);
  }
}

void DefMesh::translate(Vec3f & trans){
  for(int i=0;i<nvert;i++){
    vert[i]+=trans;
  }
}

void DefMesh::scale(Vec3f & sc){
  for(int i=0;i<nvert;i++){
    vert[i].x*=sc.x;
    vert[i].y*=sc.y;
    vert[i].z*=sc.z;
  }
}
