#ifndef VOXBIN_H
#define VOXBIN_H
#ifdef USING_GLEE
#include <GLee.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "vec.h"
#include "math.h"

class Cell{
 public:
  Vec3f * points;
  void ** userData;
  int maxpoints;
  int npoints;
  Cell(){
    userData=0;
    points=0;
    npoints=0;
    maxpoints=0;
  }
  ~Cell(){
    delete [] userData;
    delete [] points;
    maxpoints=0;
    npoints=0;
    points=0;
    userData=0;
  }
  void insert(const Vec3f & pt,void * ud){
    ensure(npoints+1);
    points[npoints]=pt;
    userData[npoints++]=ud;
  }
  void removeUserData(const Vec3f & pt){
    for(int i=0;i<npoints;i++){
      double len=(pt-points[i]).len();
      if(len==0){
	userData[i]=0;
      }
    }
  }
  int distanceQuery(const Vec3f & pt,const double rad,void ** hits,int maxhits){
    int found=0;
    for(int i=0;i<npoints;i++){
      double diff=(points[i]-pt).len();
      if(diff<rad){
	if(found<maxhits){
	  hits[found]=userData[i];
	}
	//printf("%f %f %f\n",points[i].x,points[i].y,points[i].z);
	found++;
      }
    }
    return found;
  }
  void ensure(int n){
    if(n>=maxpoints){
      int newmax=(n>(maxpoints+4))?n:maxpoints+4;
      void ** newud=new void *[newmax];
      Vec3f * newpoints = new Vec3f[newmax];
      memcpy(newud,userData,sizeof(void *)*maxpoints);
      memcpy(newpoints,points,sizeof(Vec3f)*maxpoints);
      delete [] userData;
      delete [] points;
      userData=newud;
      points=newpoints;
      maxpoints=newmax;
    }
  }
  void print(){
    printf("[");
    for(int i=0;i<npoints;i++){
      printf("(%f %f %f)",points[i].x,points[i].y,points[i].z);
    }
    printf("]");
  }
  void draw(){
    glBegin(GL_POINTS);
    for(int i=0;i<npoints;i++){
      glVertex3f(points[i].x,points[i].y,points[i].z);
    }
    glEnd();
  }
};

class Row{
 public:
      
  Cell * cells;
  Row(){
    cells=0;
  }
  ~Row(){
    delete [] cells;
  }
  void insert(const Vec3f & pt,void * ud,
	      const double *rangex,
	      const int nbins){
    if(!cells)cells = new Cell[nbins];
    int celli=(int)((nbins-1)*(pt.x-rangex[0])/(rangex[1]-rangex[0]));
    return cells[celli].insert(pt,ud);
  }
  void removeUserData(const Vec3f & pt,const double * rangex,int nbins){
    if(!cells)return;
    int celli=(int)((nbins-1)*(pt.x-rangex[0])/(rangex[1]-rangex[0]));
    return cells[celli].removeUserData(pt);
  }
  int distanceQuery(const Vec3f & pt,const double rad,
		    const double *rangex,
		    const int nbins,void ** hits,int maxhits){
    int found=0;
    if(!cells)return found;
    int mincelli=(int)((nbins-1)*(pt.x-rad-rangex[0])/(rangex[1]-rangex[0]));
    int maxcelli=(int)((nbins-1)*(pt.x+rad-rangex[0])/(rangex[1]-rangex[0]));
    mincelli=(mincelli<0)?0:mincelli;
    maxcelli=(maxcelli>=nbins)?(nbins-1):maxcelli;
    for(int i=mincelli;i<=maxcelli;i++){
      found+=cells[i].distanceQuery(pt,rad,hits+found,maxhits-found);
    }
    return found;
  }
  int bruteDistanceQuery(const Vec3f & pt,const double rad,
			 const double *rangex,
			 const int nbins){
    int found=0;
    if(!cells)return found;
    for(int i=0;i<nbins;i++){
      found+=cells[i].distanceQuery(pt,rad,0,0);
    }
    return found;
  }
  void print(int nbins){
    if(!cells)printf("[]");
    else {
      printf("[");
      for(int i=0;i<nbins;i++){
	cells[i].print();
      }
      printf("]\n");
    }
  }
  void draw(int nbins){
    if(!cells)return;
    for(int i=0;i<nbins;i++)cells[i].draw();
  }
};

class Slice{
 public:
    
  Row * rows;
  Slice(){
    rows=0;
  }
  ~Slice(){
    delete [] rows;
    rows=0;
  }
  void insert(const Vec3f & point,void * ud,
	      const double *rangex,
	      const double *rangey,
	      const int nbins){
    if(!rows)
      rows = new Row[nbins];
    int rowi=(int)((nbins-1)*(point.y-rangey[0])/(rangey[1]-rangey[0]));
    rows[rowi].insert(point,ud,rangex,nbins);
  }
  void removeUserData(const Vec3f & point,
		      const double *rangex,
		      const double *rangey,
		      const int nbins){
    if(!rows)return;
    int rowi=(int)((nbins-1)*(point.y-rangey[0])/(rangey[1]-rangey[0]));
    rows[rowi].removeUserData(point,rangex,nbins);
  }
  int distanceQuery(const Vec3f & pt,const double rad,
		    const double *rangex,
		    const double *rangey,
		    const int nbins,void ** hits,int maxhits){
    int found=0;
    if(!rows)
      return found;
    int minrowi=(int)((nbins-1)*(pt.y-rad-rangey[0])/(rangey[1]-rangey[0]));
    int maxrowi=(int)((nbins-1)*(pt.y+rad-rangey[0])/(rangey[1]-rangey[0]));
    minrowi=(minrowi<0)?0:minrowi;
    maxrowi=(maxrowi>=nbins)?(nbins-1):maxrowi;
    for(int i=minrowi;i<=maxrowi;i++){
      found+=rows[i].distanceQuery(pt,rad,rangex,nbins,hits+found,maxhits-found);
    }
    return found;
  }
  int bruteDistanceQuery(const Vec3f & pt,const double rad,
			 const double *rangex,
			 const double *rangey,
			 const int nbins){
    int found=0;
    if(!rows)
      return found;
    for(int i=0;i<nbins;i++){
      found+=rows[i].bruteDistanceQuery(pt,rad,rangex,nbins);
    }
    return found;
  }
  void print(int nbins){
    if(!rows)printf("[]\n");
    else{
      printf("[");
      for(int i=0;i<nbins;i++){
	rows[i].print(nbins);
      }
      printf("]\n");
    }
  }
  void draw(int nbins){
    if(!rows)return;
    for(int i=0;i<nbins;i++){
      rows[i].draw(nbins);
    }
  }
};

class VoxBin
{  
 public:
  int nbins;
  double rangez[2];
  double rangex[2];
  double rangey[2];
 public:
  
  friend class Row;
  friend class Cell;
  friend class Slice;
  Slice * slices;
  VoxBin(int bins=10){
    printf("resizing\n");
    slices=0;
    resize(bins);
    printf("done resizing\n");
    setRanges(0,1,0,1,0,1);
    printf("making new voxbin %d\n",bins);
  }
  void resize(int bins){
    free();
    nbins=bins;
    slices = new Slice[nbins];
  }
  void free(){
    delete [] slices;
    slices=0;
    nbins=0;
  }
  ~VoxBin(){
    free();
  }
  void setRanges(double minx,double maxx,
		 double miny,double maxy,
		 double minz,double maxz){
    rangez[0]=minz;
    rangez[1]=maxz;
    rangey[0]=miny;
    rangey[1]=maxy;
    rangex[0]=minx;
    rangex[1]=maxx;
  }
  void insert(const Vec3f & point,void * ud=0){
    int slicei=(int)((nbins-1)*(point.z-rangez[0])/(rangez[1]-rangez[0]));
    slices[slicei].insert(point,ud,rangex,rangey,nbins);
  }
  void removeUserData(const Vec3f & point){
    int slicei=(int)((nbins-1)*(point.z-rangez[0])/(rangez[1]-rangez[0]));
    slices[slicei].removeUserData(point,rangex,rangey,nbins);
  }
  int distanceQuery(const Vec3f & pt,const double rad,void ** hits=0,int maxhits=0){
    int found=0;
    int minslicei=(int)((nbins-1)*(pt.z-rad-rangez[0])/(rangez[1]-rangez[0]));
    int maxslicei=(int)((nbins-1)*(pt.z+rad-rangez[0])/(rangez[1]-rangez[0]));
    minslicei=(minslicei<0)?0:minslicei;
    maxslicei=(maxslicei>=nbins)?(nbins-1):maxslicei;
    
    for(int i=minslicei;i<=maxslicei;i++){
      found+=slices[i].distanceQuery(pt,rad,rangex,rangey,nbins,hits+found,maxhits-found);
    }
    return found;
  }
  int bruteDistanceQuery(const Vec3f & pt,const double rad){
    int found=0;
    for(int i=0;i<nbins;i++){
      found+=slices[i].bruteDistanceQuery(pt,rad,rangex,rangey,nbins);
    }
    return found;
  }
  void draw(){
    for(int i=0;i<nbins;i++){
      slices[i].draw(nbins);
    }
  }
  void print(){
    for(int i=0;i<nbins;i++){
      printf("slice %d\n",i);
      slices[i].print(nbins);
    }
  }
};


#endif
