#include "segmentdlg.h"
#include "imagedata.h"
#include "image_utils.h"
#include "timer.h"
IMPLEMENT_DYNAMIC_CLASS(SegmentDialog, wxDialog)

BEGIN_EVENT_TABLE(SegmentDialog, wxDialog)
    EVT_COMMAND_SCROLL_THUMBTRACK(ID_Thresh,SegmentDialog::OnThresh)
    EVT_BUTTON(ID_OK,SegmentDialog::OnOK)
    EVT_BUTTON(ID_Close,SegmentDialog::OnClose)
    EVT_BUTTON(ID_Load,SegmentDialog::OnLoad)
    EVT_BUTTON(ID_Add,SegmentDialog::OnAdd)
    EVT_BUTTON(ID_Remove,SegmentDialog::OnRemove)
    EVT_CLOSE(SegmentDialog::OnCloseEvent)
    EVT_IDLE(SegmentDialog::OnIdle)
END_EVENT_TABLE()

extern GLCanvas * glCanvas;
extern wxListBox * imageList;
extern wxListBox * calibImageList;
extern wxListBox * segImageList;

/**Container for a Wrect and a color region.
 * The Wrect and associated color region are used
 * for the color segmentation
 */
class RectAndRegion{
public:
  Wrect rect;
  Image8 region;
  RectAndRegion(){
  }
};

///sync the GLCanvas with the internal state.
void SegmentDialog::SyncDisplay(){
  int sel=GetSelectionSafe(calibImageList);
  if(sel>=0){
    ImageData * imd = (ImageData *)calibImageList->GetClientData(sel);
    Image8 mask=SegmentImage(imd->image);
    ImageData imdummy;
    imdummy.mask=mask;
    imdummy.image=imd->image;
    imdummy.flags=HAS_MASK;
    glCanvas->UpdateTexture(imdummy);
  }
}
/**Load button callback.
 * Allow user to select a file for use in the
 * background subtraction.
 */
void SegmentDialog::OnLoad(wxCommandEvent & event){
  wxFileDialog fd(GetParent(),wxT("Select file to load"),wxT(""),wxT(""),wxT("Images|*.ppm;*.pgm;*.tga"),wxOPEN);
  if(wxID_OK==fd.ShowModal()){
    wxArrayString names;
    wxArrayString paths;
    fd.GetFilenames(names);
    fd.GetPaths(paths);
    
    for(int i=0;i<names.GetCount();i++){
      path=paths[i];
      fileName->SetLabel(names[i]);
      bgimage.read((const char *)paths[i].ToAscii());
      printf("loading background image %s\n",(const char*)paths[i].ToAscii());
    }
  }
}
/** Add button callback.
 * Add a new region to the regions list, that 
 * currently has no color region associated with it.
 */
void SegmentDialog::OnAdd(wxCommandEvent & event){
    static int id=0;
    wxString name = wxString::Format(wxT("region-%02d"),id);
    regions->Append(name,(void *)0);
    glCanvas->RegisterListener(this);
    regions->SetSelection(regions->GetCount()-1);
    id++;
}

/** Remove button callback.
 * Remove the currently selected region from
 * the regions list.
 */
void SegmentDialog::OnRemove(wxCommandEvent & event){
    wxArrayInt sel;
    regions->GetSelections(sel);
    for(int i=sel.GetCount()-1;i>=0;i--){
	RectAndRegion * rnr=
	  (RectAndRegion *)regions->GetClientData(sel[i]);
	if(rnr)delete rnr;
	regions->Delete(sel[i]);
    }
    RegionsChanged();
}
/** threshold scroll event callback.
 * Call SyncDisplay to reflect the change.
 */
void SegmentDialog::OnThresh(wxScrollEvent & event){
  SyncDisplay();
}

/** OK callback.
 * Set the mode to segmenting.
 */
void SegmentDialog::OnOK(wxCommandEvent & event){
  wxArrayInt sel;
  calibImageList->GetSelections(sel);
  segmenting=sel.GetCount();
  SetStatusText(wxT("Segmenting"));
}

/** Idle callback.
 * If in segmenting mode, the images from the CalibImageList
 * will be segmented and placed in the SegImageList.
 */
void SegmentDialog::OnIdle(wxIdleEvent & event){
  if(segmenting){
    event.RequestMore(TRUE);
    wxArrayInt sel;
    calibImageList->GetSelections(sel);
    
    if(sel.GetCount()){
      ImageData * imd = (ImageData *)calibImageList->GetClientData(sel[0]);
      if(imd){
	imd->setMask(SegmentImage(imd->image));
      }
      printf("segmenting extr:%d intr:%d\n",
	     (imd->flags&HAS_EXTRINSIC)>0,
	     (imd->flags&HAS_INTRINSIC)>0);
      if(segImageList->GetString(0)==wxT(" ")){
	segImageList->Clear();
      }
      
      segImageList->Append(calibImageList->GetString(sel[0]),
			   imd);
      calibImageList->Delete(sel[0]);
      
      //update status
      if(segmenting>1){
	SetProgress((segmenting-sel.GetCount())/(float)(segmenting-1));
      }
      else
	SetProgress(1.0f);
      glCanvas->UpdateTexture(*imd);
      glCanvas->Refresh();
      SetChangedSinceSave(IMAGES_CHANGED);
    }
    else{
      segmenting=0;
      EndModal(wxID_OK);
      Show(FALSE);
      glCanvas->UnRegisterListener(this);
      
      SetStatusText(wxT("Segmenting Complete"));
    }
    event.RequestMore(TRUE);
  }
  else{
    event.RequestMore(FALSE);
  }
}

/**Close button callback.
 * The window is hidden, and unregistered from the GLCanvas 
 */
void SegmentDialog::OnClose(wxCommandEvent & event){
  segmenting=0;
  EndModal(wxID_CANCEL);
  Show(FALSE);
  glCanvas->UnRegisterListener(this);
}

/**Called when window is closed
   Used to be commandEvent, but something changed with version upgrade
 */
void SegmentDialog::OnCloseEvent(wxCloseEvent & event){
  wxCommandEvent ce;
  OnClose(ce);
}

/**Completes the implementation of GLMouseListener.
 * Draw any currently chosen color regions.
 * \param w width of the GLCanvas
 * \param h height of the GLCavnas
 */
void SegmentDialog::drawGL(int w,int h){
  glWidth=w;
  glHeight=h;
  if(mouseDown.x>0 && regions->GetSelection()>=0){
    glColor3f(0.0,0.0,1.0);
    glBegin(GL_LINE_LOOP);
    glVertex2f(mouseDown.x,mouseDown.y);
    glVertex2f(mouseDown.x,mouseCur.y);
    glVertex2f(mouseCur.x,mouseCur.y);
    glVertex2f(mouseCur.x,mouseDown.y);
    glEnd();
  }
  
  for(int i=0;i<regions->GetCount();i++){
    if(regions->GetClientData(i)){
      RectAndRegion * rnr = (RectAndRegion *)regions->GetClientData(i);
      Wrect * rect = &(rnr->rect);
      glColor3f(1.0,0.0,0.0);
      
      int selimage=GetSelectionSafe(calibImageList);
      double imaspect=1.33333;
      if(selimage>=0){
	ImageData * id = (ImageData *)
	  calibImageList->GetClientData(selimage);
	imaspect=((double)id->image.w)/((double)id->image.h);
      }
      double glclipx=((double)w)/((double)h);
      double glclipy=1.0;
      if(glclipx<imaspect){
	glclipx=imaspect;
	glclipy=((double)h)/((double)w)*imaspect;
      }
      glclipx*=1.05;
      glclipy*=1.05;
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      
      glTranslatef(w/2.0,h/2.0,0.0);
      glScalef(((double)imaspect*w)/glclipx,
	       ((double)h)/glclipy,1.0);
      glTranslatef(-0.5,-0.5,0.0);

      glBegin(GL_LINE_LOOP);
      glVertex2f(rect->x,rect->y);
      glVertex2f((rect->x+rect->width),(rect->y));
      glVertex2f((rect->x+rect->width),(rect->y+rect->height));
      glVertex2f((rect->x),(rect->y+rect->height));
      glEnd();

      glPopMatrix();
    }
  }
}

/** Obtain the segmentation mask for an input image.
 * The segmentation for im is obtained using the current 
 * segmentation mode.
 * \param im the image to obtain the segmentation for
 */
Image8 SegmentDialog::SegmentImage(Image8 & im){
  Image8 mask(im.width,
	      im.height,1);
  Timer t;
  t.start();
  //If the segment mode is color
  if(nb->GetSelection()==0){
    mask=255;
    for(int i=0;i<regions->GetCount();i++){
      if(regions->GetClientData(i)){
	RectAndRegion * rnr = (RectAndRegion *)regions->GetClientData(i);
	mask&=(colorSeg(im,rnr->region));
      }
    }
    mask=mask>0;//0 or 1
  }
  else{
    if(bgimage.width == im.width &&
       bgimage.height == im.height){
      Imagef diff(im.w,im.h,1);
      for(int i=0;i<im.h;i++){
	for(int j=0;j<im.w;j++){
	  diff(j,i)=fabs((float)im(j,i)-(float)bgimage(j,i));
	}
      }
      mask=diff.len()>thresh->GetValue();
    }
    else{
      printf("Background image (%d,%d) is not same dimension as input image (%d,%d)\n",bgimage.width,bgimage.height,im.width,im.height);
      
    }
  }
  mask.clean((im.width+im.height)/4);
  mask.clean((im.width+im.height)/4,0);
  mask.spur();
  printf("segmentation took in total %f\n",t.stop());

  return mask;
}

/** 
 */
void SegmentDialog::RegionsChanged(){
  SyncDisplay();
}

/** Completes the implementation of GLMouseListener
 * Currently active region in regions list is updated based
 * on the mouse drags.
 */
void SegmentDialog::mouse(wxMouseEvent & event,int GLw,int GLh){
  Raise();
    if(event.ButtonDown()){
	mouseDown=event.GetPosition();
    }
    if(event.ButtonUp() || event.Dragging()){
	mouseUp=event.GetPosition();
	
	int sel=regions->GetSelection();
	if(sel>=0){
	    double x=mouseDown.x;
	    double y=mouseDown.y;
	    double w=abs(mouseDown.x-mouseUp.x);
	    double h=abs(mouseDown.y-mouseUp.y);
	    if(x>mouseUp.x)
		x=mouseUp.x;
	    if(y>mouseUp.y)
		y=mouseUp.y;
	    int selimage=GetSelectionSafe(calibImageList);
	    if(selimage>=0){
	      ImageData * id = (ImageData *)
		calibImageList->GetClientData(selimage);
	      
	      //The transformation from screen coords to the
	      //actual image coordinates changed because of the
	      //proper aspect sizing
	      double glclipx=((double)GLw)/((double)GLh);
	      double glclipy=1.0;
	      double imaspect=((double)id->image.width)/
		((double)id->image.height);
	      if(glclipx<imaspect){
		glclipx=imaspect;
		glclipy=((double)GLh)/((double)GLw)*imaspect;
	      }
	      glclipx*=1.05;
	      glclipy*=1.05;
	      
	      x=glclipx*(x-GLw/2.0)/(((double)GLw)*imaspect)+0.5;
	      y=glclipy*(y-GLh/2.0)/(((double)GLh))+0.5;
	      w=glclipx*w/(((double)GLw)*imaspect);
	      h=glclipy*h/((double)GLh);
				
	      if(regions->GetClientData(sel)==0){
		RectAndRegion * d = new RectAndRegion();
		d->rect=Wrect(x,y,w,h);
		d->region=id->image.subimage((int)(d->rect.x*id->image.width),
					     (int)(d->rect.y*id->image.height),
					     (int)(d->rect.width*id->image.width),
					     (int)(d->rect.height*id->image.height));
		regions->SetClientData(sel,d);
	      }
	      else{
		int selimage=GetSelectionSafe(calibImageList);
		RectAndRegion * d = (RectAndRegion *)
		  regions->GetClientData(sel);
		d->rect=Wrect(x,y,w,h);
		d->region=id->image.subimage((int)(d->rect.x*id->image.width),
					     (int)(d->rect.y*id->image.height),
					     (int)(d->rect.width*id->image.width),
					     (int)(d->rect.height*id->image.height));
		printf("subimage %f %f %f %f (%d,%d)\n",d->rect.x,d->rect.y,
		       d->rect.width,d->rect.height,
		       id->image.width,id->image.height);
		regions->SetClientData(sel,d);
	      }
	      RegionsChanged();
	    }
	}
	if(event.ButtonUp())
	  mouseDown=wxPoint(-1,-1);
    }
    mouseCur=event.GetPosition();
}

/**Show/Hide the dialog.
 * On Show(Hide) the object is Registered(Unregistered)
 * with the GLCanvas.
 */
bool SegmentDialog::Show(const bool show){
  if(show){
    nb->SetSelection(0);
    if(regions->GetCount()>0){
      glCanvas->RegisterListener(this);
    }
    SyncDisplay();
  }
  else{
    if(glCanvas)
      glCanvas->UnRegisterListener(this);
  }
  wxDialog::Show(show);
  return TRUE;
}
